r"""
Discontinous Galekrin method specific terms

Note
----

In einsum calls the following convention is used:

 `i` represents iterating over all cells of a region;

 `n` represents iterating over selected cells of a region, for example
 over cells on boundary;

 `b` represents iterating over basis functions of state variable;

 `d` represents iterating over basis functions of test variable;

 `k`, `l` , `m` represent iterating over geometric dimensions, for example
 coordinates of velocity or facet normal vector or rows and columns of diffusion
 tensor;

 `q` represents iterating over quadrature points;

 `f` represents iterating over facets of cell;

"""
import numpy as nm

# sfepy imports
from sfepy.terms.terms import Term, terms
from sfepy.base.base import output


class DGTerm(Term):
    r"""
    Abstract base class for DG terms, provides alternative call_function and
    eval_real methods to accommodate returning iels and vals.
    """
    poly_space_base = "legendre"
    def call_function(self, out, fargs):
        try:
            out, status = self.function(out, *fargs)

        except (RuntimeError, ValueError):
            terms.errclear()
            raise

        if status:
            terms.errclear()
            raise ValueError('term evaluation failed! (%s)' % self.name)

        return out, status

    def eval_real(self, shape, fargs, mode='eval', term_mode=None,
                  diff_var=None, **kwargs):
        out = nm.empty(shape, dtype=nm.float64)

        if mode == 'weak':
            out, status = self.call_function(out, fargs)

        else:
            status = self.call_function(out, fargs)

        return out, status

    @staticmethod
    def _get_nbrhd_dof_indexes(cells, nrbhs, field):
        """Get indexes of DOFs for active and of active neighbouring cells

        Parameters
        ----------
        cells : array_like
            cells indicies
        nrbhs : array_like
            cell - nbrhd indicies
        field : DGField

        Returns
        -------
        iels : ndarray
            inner and outer DOF indicies, i.e. diagonal indicies and then their
            corresponding neighbour indicies

        """
        inner_iels = field.bubble_dofs
        inner_iels = nm.stack((nm.repeat(inner_iels, field.n_el_nod),
                               nm.tile(inner_iels, field.n_el_nod).flatten()),
                              axis=-1)

        outer_iels = nm.stack(
            (nm.repeat(field.bubble_dofs[cells], field.n_el_nod),
             nm.tile(field.bubble_dofs[nrbhs], field.n_el_nod).flatten()),
             axis=-1)

        iels = nm.vstack((inner_iels, outer_iels))
        return iels


class AdvectionDGFluxTerm(DGTerm):
    r"""
    Lax-Friedrichs flux term for advection of scalar quantity :math:`p` with the
    advection velocity :math:`\ul{a}` given as a material parameter (a known
    function of space and time).

    :Definition:

    .. math::
         \int_{\partial{T_K}} \ul{n} \cdot \ul{f}^{*} (p_{in}, p_{out})q

    where

    .. math::
        \ul{f}^{*}(p_{in}, p_{out}) =  \ul{a}  \frac{p_{in} + p_{out}}{2}  +
        (1 - \alpha) \ul{n} C \frac{ p_{in} - p_{out}}{2},

    :math:`\alpha \in [0, 1]`; :math:`\alpha = 0` for upwind scheme,
    :math:`\alpha = 1` for central scheme,  and

    .. math::
        C = \max_{p \in [?, ?]}\left\lvert n_x a_1 +
        n_y a_2 \right\rvert =
        \max_{p \in [?, ?]} \left\lvert  \ul{n} \cdot \ul{a} \right\rvert


    the :math:`p_{in}` resp. :math:`p_{out}`
    is solution on the boundary of the element
    provided by element itself resp. its
    neighbor and :math:`\ul{a}` is advection velocity.

    :Arguments 1:
        - material : :math:`\ul{a}`
        - virtual  : :math:`q`
        - state    : :math:`p`

    :Arguments 3:
        - material    : :math:`\ul{a}`
        - virtual     : :math:`q`
        - state       : :math:`p`
        - opt_material : :math:`\alpha`
    """

    alpha = 0
    name = "dw_dg_advect_laxfrie_flux"
    modes = ("weak",)
    arg_types = ('opt_material', 'material_advelo', 'virtual', 'state')
    arg_shapes = [{'opt_material'   : '.: 1',
                   'material_advelo': 'D, 1',
                   'virtual'         : (1, 'state'),
                   'state'           : 1
                   },
                  {'opt_material': None}]
    integration = 'volume'
    symbolic = {'expression': 'div(a*p)*w',
                'map' : {'p': 'state', 'a': 'material', 'v': 'virtual'}
                }

    def get_fargs(self, alpha, advelo, test, state,
                  mode=None, term_mode=None, diff_var=None, **kwargs):

        if alpha is not None:
            self.alpha = alpha

        field = state.field
        region = field.region

        if "DG" not in field.family_name:
            raise ValueError("Used DG term with non DG field {} of family {}"
                             .format(field.name, field.family_name))

        fargs = (state, diff_var, field, region, advelo[:, 0, :, 0])
        return fargs

    def function(self, out, state, diff_var, field, region, advelo):

        if diff_var is not None:

            fc_n = field.get_cell_normals_per_facet(region)
            C = nm.abs(nm.einsum("ifk,ik->if", fc_n, advelo))

            nbrhd_idx = field.get_facet_neighbor_idx(region, state.eq_map)
            active_cells, active_facets = nm.where(nbrhd_idx[:, :, 0] >= 0)
            active_nrbhs = nbrhd_idx[active_cells, active_facets, 0]

            in_fc_b, out_fc_b, whs = field.get_both_facet_base_vals(state, region)

            # TODO broadcast advelo to facets
            #  - maybe somehow get values of advelo at them?
            # compute values
            inner_diff = nm.einsum("nfk, nfk->nf",
                                   fc_n,
                                   advelo[:, None, :]
                                   + nm.einsum("nfk, nf->nfk",
                                               (1 - self.alpha) * fc_n, C)) / 2.
            outer_diff = nm.einsum("nfk, nfk->nf",
                                   fc_n,
                                   advelo[:, None, :]
                                   - nm.einsum("nfk, nf->nfk",
                                               (1 - self.alpha) * fc_n, C)) / 2.

            inner_vals = nm.einsum("nf, ndfq, nbfq, nfq -> ndb",
                                   inner_diff,
                                   in_fc_b,
                                   in_fc_b,
                                   whs)
            outer_vals = nm.einsum("i, idq, ibq, iq -> idb",
                                   outer_diff[active_cells, active_facets],
                                   in_fc_b[active_cells, :, active_facets],
                                   out_fc_b[active_cells, :, active_facets],
                                   whs[active_cells, active_facets])

            vals = nm.vstack((inner_vals, outer_vals))
            vals = vals.flatten()

            # compute positions within matrix
            iels = self._get_nbrhd_dof_indexes(active_cells, active_nrbhs, field)

            out = (vals, iels[:, 0], iels[:, 1], state, state)
        else:
            fc_n = field.get_cell_normals_per_facet(region)
            # get maximal wave speeds at facets
            C = nm.abs(nm.einsum("ifk,ik->if", fc_n, advelo))

            facet_base_vals = field.get_facet_base(base_only=True)
            in_fc_v, out_fc_v, weights = field.get_both_facet_state_vals(state,
                                                                         region)
            # get sane facet base shape
            fc_b = facet_base_vals[:, 0, :, 0, :].T
            # (n_el_nod, n_el_facet, n_qp)

            fc_v_avg = (in_fc_v + out_fc_v)/2.
            fc_v_jmp = in_fc_v - out_fc_v

            central = nm.einsum("ik,ifq->ifkq", advelo, fc_v_avg)
            upwind = (1 - self.alpha) / 2. * nm.einsum("if,ifk,ifq->ifkq",
                                                       C, fc_n, fc_v_jmp)

            cell_fluxes = nm.einsum("ifk,ifkq,dfq,ifq->id",
                                    fc_n, central + upwind, fc_b, weights)

            out[:] = 0.0
            n_el_nod = field.n_el_nod
            for i in range(n_el_nod):
                out[:, :, i, 0] = cell_fluxes[:, i, None]

        status = None
        return out, status


class DiffusionDGFluxTerm(DGTerm):
    r"""
    Basic DG diffusion flux term for scalar quantity.

     :Definition:

    .. math::
         \int_{\partial{T_K}} D \langle \nabla p \rangle [q] \mbox{ , }
         \int_{\partial{T_K}} D \langle \nabla q \rangle [p]

    where

    .. math::
        \langle \nabla \phi \rangle = \frac{\nabla\phi_{in} + \nabla\phi_{out}}{2}

    .. math::
        [\phi] = \phi_{in} - \phi_{out}
    :math:


    The :math:`p_{in}` resp. :math:`p_{out}`
    is solution on the boundary of the element
    provided by element itself resp. its neighbour.

    :Arguments 1:
        - material    : :math:`D`
        - state       : :math:`p`
        - virtual     : :math:`q`

    :Arguments 2:
        - material : :math:`D`
        - virtual  : :math:`q`
        - state    : :math:`p`


    """
    name = "dw_dg_diffusion_flux"
    arg_types = (('material', 'state', 'virtual'), # left
                 ('material', 'virtual', 'state') # right
                 )
    arg_shapes = [{'material': '1, 1',
                   'virtual/avg_state': (1, None),
                   'state/avg_state' : 1,
                   'virtual/avg_virtual': (1, None),
                   'state/avg_virtual' : 1,
                   }]
    integration = 'volume'
    modes = ('avg_state', 'avg_virtual')

    def get_fargs(self, diff_tensor, test, state,
                  mode=None, term_mode=None, diff_var=None, **kwargs):

        field = state.field
        region = field.region

        if "DG" not in field.family_name:
            raise ValueError("Used DG term with non DG field {} of family {}"
                             .format(field.name, field.family_name))
        if self.mode == "avg_state":
            # put state where it is expected by the function
            state = test
        fargs = (state, diff_var, field, region, diff_tensor[:, 0, :, :])
        return fargs

    def function(self, out, state, diff_var, field, region, D):
        if diff_var is not None:  # matrix mode
            # outR = out.copy()[..., 0:1]
            out = self._function_matrix(out, state, diff_var, field, region, D)

            # vals, ielsi, ielsj = out[:3]
            # from scipy.sparse import coo_matrix
            # extra = coo_matrix((vals, (ielsi, ielsj)),
            #                    shape=2*(field.n_el_nod * field.n_cell,))
            # M = extra.toarray()
            # u = state.data[0]
            # Mu = nm.dot(M, u).reshape((field.n_cell, field.n_el_nod))
            #
            # outR = self._function_residual(outR, state, diff_var, field,
            #                                   region, D).squeeze()
            # from matplotlib import pyplot as plt
            # plt.imshow((Mu - outR).T, aspect="auto")
            # plt.colorbar()
            #
            # nbrhd_idx = field.get_facet_neighbor_idx(region, state.eq_map)
            # bcells = nm.where(nbrhd_idx[:, :, 0] < 0)[0],
            # plt.vlines(bcells, -.5, 15, alpha=.3)
            # plt.show()

        else:  # residual mode
            out = self._function_residual(out, state, diff_var, field, region, D)

        status = None
        return out, status


    def _function_matrix(self, out, state, diff_var, field, region, D):
        fc_n = field.get_cell_normals_per_facet(region)

        nbrhd_idx = field.get_facet_neighbor_idx(region, state.eq_map)
        active_cells, active_facets = nm.where(nbrhd_idx[:, :, 0] >= 0)
        active_nrbhs = nbrhd_idx[active_cells, active_facets, 0]

        inner_facet_base, outer_facet_base, whs = \
            field.get_both_facet_base_vals(state, region, derivative=False)

        inner_facet_base_d, outer_facet_base_d, _ = \
            field.get_both_facet_base_vals(state, region, derivative=True)

        if self.mode == 'avg_state':
            # content of diagonal
            inner_vals = nm.einsum("nkl, nbfkq, nfk, ndfq, nfq->ndb",
                                   D,
                                   inner_facet_base_d / 2,  # state
                                   fc_n,
                                   inner_facet_base,  # test
                                   whs)
            outer_vals = nm.einsum(
                "ikl, ibkq, ik, idq, iq->idb",
                D[active_cells],
                outer_facet_base_d[active_cells, :, active_facets] / 2,  # state
                fc_n[active_cells, active_facets],
                inner_facet_base[active_cells, :, active_facets],  # test
                whs[active_cells, active_facets])

        elif self.mode == 'avg_virtual':
            # content of diagonal
            inner_vals = nm.einsum("nkl, ndfkq, nfk, nbfq, nfq->ndb",
                                   D,
                                   inner_facet_base_d / 2,  # test
                                   fc_n,
                                   inner_facet_base,  # state
                                   whs)

            outer_vals = nm.einsum("ikl, idkq, ik, ibq, iq->idb",
                   D[active_cells],
                   inner_facet_base_d[active_cells, :, active_facets] / 2,  # test
                   fc_n[active_cells, active_facets],
                   - outer_facet_base[active_cells, :, active_facets],  # state
                   whs[active_cells, active_facets])

        iels = self._get_nbrhd_dof_indexes(active_cells, active_nrbhs, field)

        vals = nm.vstack((inner_vals, outer_vals))
        vals = vals.flatten()

        #               i           j
        out = (vals, iels[:, 0], iels[:, 1], state, state)
        return out

    def _function_residual(self, out, state, diff_var, field, region, D):
        fc_n = field.get_cell_normals_per_facet(region)

        # get base values
        inner_facet_base, outer_facet_base, _ = \
            field.get_both_facet_base_vals(state, region, derivative=False)
        inner_facet_base_d, outer_facet_base_d, _ = \
            field.get_both_facet_base_vals(state, region, derivative=True)

        # get state values
        inner_facet_state_d, outer_facet_state_d, _ = \
            field.get_both_facet_state_vals(state, region, derivative=True)
        inner_facet_state, outer_facet_state, weights = \
            field.get_both_facet_state_vals(state, region, derivative=False)

        if self.mode == 'avg_state':
            avgDdState = (nm.einsum("ikl,ifkq->ifkq",
                                    D, inner_facet_state_d) +
                          nm.einsum("ikl,ifkq                ->ifkq",
                                    D,   outer_facet_state_d)) / 2.
            # outer_facet_base is in DG zero - hence the jump is inner value
            jmpBase = inner_facet_base

            cell_fluxes = nm.einsum("ifkq ,     ifk,  idfq,    ifq -> id",
                             avgDdState, fc_n, jmpBase, weights)

        elif self.mode == 'avg_virtual':
            avgDdbase = (nm.einsum("ikl,idfkq->idfkq",
                                   D, inner_facet_base_d)) / 2.

            jmpState = inner_facet_state - outer_facet_state
            cell_fluxes = nm.einsum("idfkq, ifk, ifq , ifq -> id",
                             avgDdbase, fc_n, jmpState, weights)

        out[:] = 0.0
        n_el_nod = field.n_el_nod
        for i in range(n_el_nod):
            out[:, :, i, 0] = cell_fluxes[:, i, None]
        return out


class DiffusionInteriorPenaltyTerm(DGTerm):
    r"""
    Penalty term used to counteract discontinuity arising when
    modeling diffusion using Discontinuous Galerkin schemes.

     :Definition:

    .. math::
         \int_{\partial{T_K}} \bar{D} C_w \frac{Ord^2}{d(\partial{T_K})}[p][q]

    where

    .. math::
        [\phi] = \phi_{in} - \phi_{out}
    :math:

    the :math:`p_{in}` resp. :math:`p_{out}`
    is solution on the boundary of the element
    provided by element itself resp. its neighbour.

    :Arguments:
        - material    : :math:`D`
        - material    : :math:`C_w`
        - state       : :math:`p`
        - virtual     : :math:`q`
    """
    name = "dw_dg_interior_penalty"
    modes = ("weak",)
    arg_types = ('material', 'material_Cw',
                 'virtual', 'state')
    arg_shapes = [{'material': '1, 1',
                   'material_Cw': '.: 1',
                   'virtual'    : (1, 'state'),
                   'state'      : 1
                   }]

    def get_fargs(self, diff_tensor, Cw, test, state, mode=None,
                  term_mode=None, diff_var=None, **kwargs):

        field = state.field
        region = field.region

        if "DG" not in field.family_name:
            raise ValueError("Used DG term with non DG field {} of family {}"
                             .format(field.name, field.family_name))

        fargs = (state, diff_var, field, region, Cw,  diff_tensor[:, 0, :, :])
        return fargs

    def function(self, out, state, diff_var, field, region, Cw, diff_tensor):

        approx_order = field.approx_order

        inner_facet_base, outer_facet_base, whs = \
            field.get_both_facet_base_vals(state, region, derivative=False)
        facet_vols = nm.sum(whs, axis=-1)

        # nu characterizes diffusion tensor, so far we use diagonal average
        nu = nm.trace(diff_tensor, axis1=-2, axis2=-1)[..., None] / \
             diff_tensor.shape[1]
        sigma = nu * Cw * approx_order ** 2 / facet_vols

        if diff_var is not None:

            nbrhd_idx = field.get_facet_neighbor_idx(region, state.eq_map)
            active_cells, active_facets = nm.where(nbrhd_idx[:, :, 0] >= 0)
            active_nrbhs = nbrhd_idx[active_cells, active_facets, 0]

            inner = nm.einsum("nf, ndfq, nbfq, nfq -> ndb",
                               sigma,
                               inner_facet_base,  # test
                               inner_facet_base,  # state
                               whs)

            outer = nm.einsum("i, idq, ibq, iq -> idb",
                               sigma[active_cells, active_facets],
                               inner_facet_base[active_cells, :, active_facets],  # test
                               - outer_facet_base[active_cells, :, active_facets],  # state
                               whs[active_cells, active_facets])

            vals = nm.vstack((inner, outer))
            vals = vals.flatten()

            iels = self._get_nbrhd_dof_indexes(active_cells, active_nrbhs, field)

            out = (vals, iels[:, 0], iels[:, 1], state, state)

        else:
            inner_facet_state, outer_facet_state, whs = \
                field.get_both_facet_state_vals(state, region,
                                                derivative=False
                                                )
            inner_facet_base, outer_facet_base, _ = \
                field.get_both_facet_base_vals(state, region,
                                               derivative=False
                                               )
            jmp_state = inner_facet_state - outer_facet_state
            jmp_base = inner_facet_base  # - outer_facet_base

            n_el_nod = nm.shape(inner_facet_base)[1]
            cell_penalty = nm.einsum("nf,nfq,ndfq,nfq->nd",
                                     sigma, jmp_state, jmp_base, whs)

            out[:] = 0.0
            for i in range(n_el_nod):
                out[:, :, i, 0] = cell_penalty[:, i, None]

        status = None
        return out, status


class NonlinearHyperbolicDGFluxTerm(DGTerm):
    r"""
     Lax-Friedrichs flux term for nonlinear hyperpolic term of scalar quantity
     :math:`p` with the vector function :math:`\ul{f}` given as a material
     parameter.

    :Definition:

    .. math::
         \int_{\partial{T_K}} \ul{n} \cdot f^{*} (p_{in}, p_{out})q

    where

    .. math::
        \ul{f}^{*}(p_{in}, p_{out}) =  \frac{\ul{f}(p_{in})
            + \ul{f}(p_{out})}{2}  +
        (1 - \alpha) \ul{n} C \frac{ p_{in} - p_{out}}{2},

    :math:`\alpha \in [0, 1]`; :math:`\alpha = 0` for upwind scheme,
    :math:`\alpha = 1` for central scheme,  and

    .. math::
        C =
        \max_{p \in [?, ?]}\left\lvert
        n_x \frac{d f_1}{d p} + n_y \frac{d f_2}{d p}
        + \cdots
        \right\rvert =

        \max_{p \in [?, ?]} \left\lvert
        \vec{n}\cdot\frac{d\ul{f}}{dp}(p)
         \right\rvert


    the :math:`p_{in}` resp. :math:`p_{out}`
    is solution on the boundary of the element
    provided by element itself resp. its
    neighbor.

    :Arguments 1:
        - material : :math:`\ul{f}`
        - material : :math:`\frac{d\ul{f}}{d p}`
        - virtual  : :math:`q`
        - state    : :math:`p`

    :Arguments 3:
        - material    : :math:`\ul{f}`
        - material    : :math:`\frac{d\ul{f}}{d p}`
        - virtual     : :math:`q`
        - state       : :math:`p`
        - opt_material : :math:`\alpha`
    """
    alf = 0
    name = "dw_dg_nonlinear_laxfrie_flux"
    modes = ("weak",)
    arg_types = ('opt_material', 'fun', 'fun_d', 'virtual', 'state')
    arg_shapes = [{'opt_material'  : '.: 1',
                   'material_fun'  : '.: 1',
                   'material_fun_d': '.: 1',
                   'virtual'       : (1, 'state'),
                   'state'         : 1
                   },
                  {'opt_material': None}]
    integration = 'volume'
    symbolic = {'expression': 'div(f(p))*w',
                'map'       : {'p': 'state', 'v': 'virtual', 'f': 'function'}
                }

    def get_fargs(self, alpha, fun, dfun, test, state,
                  mode=None, term_mode=None, diff_var=None, **kwargs):

        if alpha is not None:
            self.alf = nm.max(alpha)  # extract alpha value regardless of shape

        self.fun = fun
        self.dfun = dfun

        if diff_var is not None:
            output("Diff var is not None in nonlinear, residual only " +
                   "term"" {} !  Skipping.".format(self.name))
            return None, None, None, 0, 0
        else:
            field = state.field
            region = field.region

            if "DG" not in field.family_name:
                raise ValueError(
                        "Used DG term with non DG field {} of family {}!"
                        .format(field.name, field.family_name))

            fargs = (state, field, region, fun, dfun)
            return fargs

    def function(self, out, state, field, region, f, df):
        if state is None:
            out[:] = 0.0
            return None

        fc_n = field.get_cell_normals_per_facet(region)
        facet_base_vals = field.get_facet_base(base_only=True)
        in_fc_v, out_fc_v, weights = field.get_both_facet_state_vals(state, region)

        fc_b = facet_base_vals[:, 0, :, 0, :].T  # (n_el_nod, n_el_facet, n_qp)

        n_el_nod = field.n_el_nod

        # get maximal wave speeds at facets
        df_in = df(in_fc_v)
        df_out = df(out_fc_v)
        fc_n__dot__df_in = nm.einsum("ifk,ifqk->ifq", fc_n, df_in)
        fc_n__dot__df_out = nm.einsum("ifk,ifqk->ifq", fc_n, df_out)
        dfdn = nm.stack((fc_n__dot__df_in, fc_n__dot__df_out), axis=-1)
        C = nm.amax(nm.abs(dfdn), axis=(-2, -1))

        fc_f_avg = (f(in_fc_v) + f(out_fc_v)) / 2.
        fc_v_jmp = in_fc_v - out_fc_v

        central = fc_f_avg
        upwind = (1 - self.alf) / 2. * nm.einsum("if,ifk,ifq->ifqk",
                                                 C, fc_n, fc_v_jmp)

        cell_fluxes = nm.einsum("ifk,ifqk,dfq,ifq->id",
                                fc_n, central + upwind, fc_b, weights)

        out[:] = 0.0
        for i in range(n_el_nod):
            out[:, :, i, 0] = cell_fluxes[:, i, None]

        status = None
        return out, status


from sfepy.linalg import dot_sequences


class NonlinearScalarDotGradTerm(Term):
    r"""
     Product of virtual and divergence of vector function of state or volume dot
     product of vector function of state and gradient of scalar virtual.

    :Definition:

    .. math::
        \int_{\Omega} q \cdot \nabla \cdot \ul{f}(p) = \int_{\Omega} q \cdot
           \text{div} \ul{f}(p)  \mbox{ , }
        \int_{\Omega} \ul{f}(p) \cdot \nabla q

    :Arguments 1:
        - function : :math:`\ul{f}`
        - virtual  : :math:`q`
        - state    : :math:`p`

    :Arguments 2:
        - function : :math:`\ul{f}`
        - state    : :math:`p`
        - virtual  : :math:`q`

    TODO maybe this term would fit better to terms_dot?
    """
    name = 'dw_ns_dot_grad_s'
    arg_types = (('fun', 'fun_d', 'virtual', 'state'),
                 ('fun', 'fun_d', 'state', 'virtual'))
    arg_shapes = [{'material_fun'        : '.: 1',
                   'material_fun_d'      : '.: 1',
                   'virtual/grad_state'  : (1, None),
                   'state/grad_state'    : 1,
                   'virtual/grad_virtual': (1, None),
                   'state/grad_virtual'  : 1}]
    modes = ('grad_state', 'grad_virtual')

    @staticmethod
    def function(out, out_qp, geo, fmode):
        status = geo.integrate(out, out_qp)
        return status

    def get_fargs(self, fun, dfun, var1, var2,
                  mode=None, term_mode=None, diff_var=None, **kwargs):
        vg1, _ = self.get_mapping(var1)
        vg2, _ = self.get_mapping(var2)

        if diff_var is None:
            if self.mode == 'grad_state':
                # TODO rewrite using einsum?
                geo = vg1
                bf_t = vg1.bf.transpose((0, 1, 3, 2))
                val_grad_qp = self.get(var2, 'grad')
                out_qp = dot_sequences(bf_t, val_grad_qp, 'ATB')

            else:
                geo = vg2
                val_qp = fun(self.get(var1, 'val'))[..., 0, :].swapaxes(-2, -1)
                out_qp = dot_sequences(vg2.bfg, val_qp, 'ATB')

            fmode = 0

        else:
            raise ValueError("Matrix mode not supported for {}"
                             .format(self.name))
            # however it could be with use of dfun
            if self.mode == 'grad_state':
                geo = vg1
                bf_t = vg1.bf.transpose((0, 1, 3, 2))
                out_qp = dot_sequences(bf_t, vg2.bfg, 'ATB')

            else:
                geo = vg2
                out_qp = dot_sequences(vg2.bfg, vg1.bf, 'ATB')

            fmode = 1

        return out_qp, geo, fmode



import numpy as nm

from sfepy.linalg import dot_sequences
from sfepy.terms.terms import Term, terms
from sfepy.terms.terms_th import THTerm, ETHTerm
from sfepy.terms.terms_elastic import CauchyStressTerm

class BiotTerm(Term):
    r"""
    Biot coupling term with :math:`\alpha_{ij}` given in:

    * vector form exploiting symmetry - in 3D it has the
      indices ordered as :math:`[11, 22, 33, 12, 13, 23]`, in 2D it has
      the indices ordered as :math:`[11, 22, 12]`,

    * matrix form - non-symmetric coupling parameter.

    Corresponds to weak forms of Biot gradient and divergence terms.
    Can be evaluated. Can use derivatives.

    :Definition:

    .. math::
        \int_{\Omega}  p\ \alpha_{ij} e_{ij}(\ul{v}) \mbox{ , } \int_{\Omega}
        q\ \alpha_{ij} e_{ij}(\ul{u})

    :Arguments 1:
        - material : :math:`\alpha_{ij}`
        - virtual  : :math:`\ul{v}`
        - state    : :math:`p`

    :Arguments 2:
        - material : :math:`\alpha_{ij}`
        - state    : :math:`\ul{u}`
        - virtual  : :math:`q`

    :Arguments 3:
        - material    : :math:`\alpha_{ij}`
        - parameter_v : :math:`\ul{u}`
        - parameter_s : :math:`p`
    """
    name = 'dw_biot'
    arg_types = (('material', 'virtual', 'state'),
                 ('material', 'state', 'virtual'),
                 ('material', 'parameter_v', 'parameter_s'))
    arg_shapes = [{'material' : 'S, 1',
                  'virtual/grad' : ('D', None), 'state/grad' : 1,
                  'virtual/div' : (1, None), 'state/div' : 'D',
                  'parameter_v' : 'D', 'parameter_s' : 1},
                  {'material' : 'D, D'}]

    modes = ('grad', 'div', 'eval')

    def get_fargs(self, mat, vvar, svar,
                  mode=None, term_mode=None, diff_var=None, **kwargs):

        sym_mode = False if mat.shape[-2] == mat.shape[-1] > 1 else True
        if not sym_mode:
            sh = mat.shape
            # the gradient given by 'self.get' is transposed
            mat = nm.swapaxes(mat, 2, 3)
            mat = mat.reshape(sh[:2] + (sh[2]**2, 1))

        if self.mode == 'grad':
            qp_var, qp_name = svar, 'val'

        else:
            if sym_mode:
                qp_var, qp_name = vvar, 'cauchy_strain'
            else:
                qp_var, qp_name = vvar, 'grad'

        if mode == 'weak':
            vvg, _ = self.get_mapping(vvar)
            svg, _ = self.get_mapping(svar)

            if diff_var is None:
                val_qp = self.get(qp_var, qp_name)
                if qp_name == 'grad':
                    sh = val_qp.shape
                    val_qp = val_qp.reshape(sh[:2] + (sh[2]**2, 1))

                fmode = 0

            else:
                val_qp = nm.array([0], ndmin=4, dtype=nm.float64)
                fmode = 1

            return 1.0, val_qp, mat, svg, vvg, fmode

        elif mode == 'eval':
            vvg, _ = self.get_mapping(vvar)

            if sym_mode:
                strain = self.get(vvar, 'cauchy_strain')
            else:
                strain = self.get(vvar, 'grad')
                sh = strain.shape
                strain = strain.reshape(sh[:2] + (sh[2]**2, 1))

            pval = self.get(svar, 'val')

            return 1.0, pval, strain, mat, vvg

        else:
            raise ValueError('unsupported evaluation mode in %s! (%s)'
                             % (self.name, mode))

    def get_eval_shape(self, mat, vvar, svar,
                       mode=None, term_mode=None, diff_var=None, **kwargs):
        n_el, n_qp, dim, n_en, n_c = self.get_data_shape(vvar)

        return (n_el, 1, 1, 1), vvar.dtype

    def set_arg_types(self):
        self.function = {
            'grad' : terms.dw_biot_grad,
            'div' : terms.dw_biot_div,
            'eval' : terms.d_biot_div,
        }[self.mode]

class BiotStressTerm(CauchyStressTerm):
    r"""
    Evaluate Biot stress tensor.

    It is given in the usual vector form exploiting symmetry: in 3D it has 6
    components with the indices ordered as :math:`[11, 22, 33, 12, 13, 23]`, in
    2D it has 3 components with the indices ordered as :math:`[11, 22, 12]`.

    Supports 'eval', 'el_avg' and 'qp' evaluation modes.

    :Definition:

    .. math::
        - \int_{\Omega} \alpha_{ij} \bar{p}

    .. math::
        \mbox{vector for } K \from \Ical_h:
        - \int_{T_K} \alpha_{ij} \bar{p} / \int_{T_K} 1

    .. math::
        - \alpha_{ij} \bar{p}|_{qp}

    :Arguments:
        - material  : :math:`\alpha_{ij}`
        - parameter : :math:`\bar{p}`
    """
    name = 'ev_biot_stress'
    arg_types = ('material', 'parameter')
    arg_shapes = {'material' : 'S, 1', 'parameter' : 1}

    @staticmethod
    def function(out, val_qp, mat, vg, fmode):
        if fmode == 2:
            out[:] = dot_sequences(mat, val_qp)
            status = 0

        else:
            status = terms.de_cauchy_stress(out, val_qp, mat, vg, fmode)

        out *= -1.0

        return status

    def get_fargs(self, mat, parameter,
                  mode=None, term_mode=None, diff_var=None, **kwargs):
        vg, _ = self.get_mapping(parameter)

        val_qp = self.get(parameter, 'val')

        fmode = {'eval' : 0, 'el_avg' : 1, 'qp' : 2}.get(mode, 1)

        return val_qp, mat, vg, fmode

class BiotTHTerm(BiotTerm, THTerm):
    r"""
    Fading memory Biot term. Can use derivatives.

    :Definition:

    .. math::
        \begin{array}{l}
        \int_{\Omega} \left [\int_0^t \alpha_{ij}(t-\tau)\,p(\tau)) \difd{\tau}
        \right]\,e_{ij}(\ul{v}) \mbox{ ,} \\
        \int_{\Omega} \left [\int_0^t
        \alpha_{ij}(t-\tau) e_{kl}(\ul{u}(\tau)) \difd{\tau} \right] q
        \end{array}

    :Arguments 1:
        - ts       : :class:`TimeStepper` instance
        - material : :math:`\alpha_{ij}(\tau)`
        - virtual  : :math:`\ul{v}`
        - state    : :math:`p`

    :Arguments 2:
        - ts       : :class:`TimeStepper` instance
        - material : :math:`\alpha_{ij}(\tau)`
        - state    : :math:`\ul{u}`
        - virtual  : :math:`q`
    """
    name = 'dw_biot_th'
    arg_types = (('ts', 'material', 'virtual', 'state'),
                 ('ts', 'material', 'state', 'virtual'))
    arg_shapes = {'material' : '.: N, S, 1',
                  'virtual/grad' : ('D', None), 'state/grad' : 1,
                  'virtual/div' : (1, None), 'state/div' : 'D'}
    modes = ('grad', 'div')

    def get_fargs(self, ts, mats, vvar, svar,
                  mode=None, term_mode=None, diff_var=None, **kwargs):
        if self.mode == 'grad':
            qp_var, qp_name = svar, 'val'

        else:
            qp_var, qp_name = vvar, 'cauchy_strain'

        n_el, n_qp, dim, n_en, n_c = self.get_data_shape(svar)

        if mode == 'weak':
            vvg, _ = self.get_mapping(vvar)
            svg, _ = self.get_mapping(svar)

            if diff_var is None:
                def iter_kernel():
                    for ii, mat in enumerate(mats):
                        val_qp = self.get(qp_var, qp_name, step=-ii)
                        mat = nm.tile(mat, (n_el, n_qp, 1, 1))
                        yield ii, (ts.dt, val_qp, mat, svg, vvg, 0)
                fargs = iter_kernel

            else:
                val_qp = nm.array([0], ndmin=4, dtype=nm.float64)
                mat = nm.tile(mats[0], (n_el, n_qp, 1, 1))
                fargs = ts.dt, val_qp, mat, svg, vvg, 1

            return fargs

        else:
            raise ValueError('unsupported evaluation mode in %s! (%s)'
                             % (self.name, mode))

class BiotETHTerm(BiotTerm, ETHTerm):
    r"""
    This term has the same definition as dw_biot_th, but assumes an
    exponential approximation of the convolution kernel resulting in much
    higher efficiency. Can use derivatives.

    :Definition:

    .. math::
        \begin{array}{l}
        \int_{\Omega} \left [\int_0^t \alpha_{ij}(t-\tau)\,p(\tau)) \difd{\tau}
        \right]\,e_{ij}(\ul{v}) \mbox{ ,} \\
        \int_{\Omega} \left [\int_0^t
        \alpha_{ij}(t-\tau) e_{kl}(\ul{u}(\tau)) \difd{\tau} \right] q
        \end{array}

    :Arguments 1:
        - ts         : :class:`TimeStepper` instance
        - material_0 : :math:`\alpha_{ij}(0)`
        - material_1 : :math:`\exp(-\lambda \Delta t)` (decay at :math:`t_1`)
        - virtual    : :math:`\ul{v}`
        - state      : :math:`p`

    :Arguments 2:
        - ts         : :class:`TimeStepper` instance
        - material_0 : :math:`\alpha_{ij}(0)`
        - material_1 : :math:`\exp(-\lambda \Delta t)` (decay at :math:`t_1`)
        - state      : :math:`\ul{u}`
        - virtual    : :math:`q`
    """
    name = 'dw_biot_eth'
    arg_types = (('ts', 'material_0', 'material_1', 'virtual', 'state'),
                 ('ts', 'material_0', 'material_1', 'state', 'virtual'))
    arg_shapes = {'material_0' : 'S, 1', 'material_1' : '1, 1',
                  'virtual/grad' : ('D', None), 'state/grad' : 1,
                  'virtual/div' : (1, None), 'state/div' : 'D'}
    modes = ('grad', 'div')

    def get_fargs(self, ts, mat0, mat1, vvar, svar,
                  mode=None, term_mode=None, diff_var=None, **kwargs):
        if self.mode == 'grad':
            qp_var, qp_name, iv = svar, 'val', 4

        else:
            qp_var, qp_name, iv = vvar, 'cauchy_strain', 3

        if mode == 'weak':
            vvg, _, key = self.get_mapping(vvar, return_key=True)
            svg, _ = self.get_mapping(svar)

            if diff_var is None:
                val_qp = self.get(qp_var, qp_name)

                key += tuple(self.arg_names[ii] for ii in [1, 2, iv])
                data = self.get_eth_data(key, qp_var, mat1, val_qp)

                val = data.history + data.values
                fargs = (ts.dt, val, mat0, svg, vvg, 0)

            else:
                val_qp = nm.array([0], ndmin=4, dtype=nm.float64)
                fargs = (ts.dt, val_qp, mat0, svg, vvg, 1)

            return fargs

        else:
            raise ValueError('unsupported evaluation mode in %s! (%s)'
                             % (self.name, mode))



from __future__ import absolute_import
import re
from copy import copy

import numpy as nm

from sfepy.base.base import assert_, Struct
import six
from six.moves import range

_depends = re.compile('r\.([a-zA-Z_\-0-9.]+)').findall

def get_parents(selector):
    """
    Given a region selector, return names of regions it is based on.
    """
    parents = _depends(selector)

    return parents

def get_dependency_graph(region_defs):
    """
    Return a dependency graph and a name-sort name mapping for given
    region definitions.
    """
    graph = {}
    name_to_sort_name = {}
    for sort_name, rdef in six.iteritems(region_defs):
        name, sel = rdef.name, rdef.select
        if name in name_to_sort_name:
            msg = 'region %s/%s already defined!' % (sort_name, name)
            raise ValueError(msg)
        name_to_sort_name[name] = sort_name

        if name not in graph:
            graph[name] = [0]

        for parent in get_parents(sel):
            graph[name].append(parent)

        if rdef.get('parent', None) is not None:
            graph[name].append(rdef.parent)

    return graph, name_to_sort_name

def sort_by_dependency(graph):
    out = []

    n_nod = len(graph)
    idone = 0
    idone0 = -1
    while idone < n_nod:

        dep_removed = 0
        for node, deps in six.iteritems(graph):

            if (len(deps) == 1) and not deps[0]:
                out.append(node)
                deps[0] = 1
                idone += 1

            elif not deps[0]:

                for ii, dep in enumerate(deps[1:]):
                    if not dep in graph:
                        msg = 'dependency %s of region %s does not exist!'
                        raise ValueError(msg % (dep, node))

                    if graph[dep][0]:
                        ir = deps.index(dep)
                        deps.pop(ir)
                        dep_removed += 1

        if (idone <= idone0) and not dep_removed:
            raise ValueError('circular dependency')
        idone0 = idone

    return out

def are_disjoint(r1, r2):
    """
    Check if the regions `r1` and `r2` are disjoint.

    Uses vertices for the check - `*_only` regions not allowed.
    """
    return len(nm.intersect1d(r1.vertices, r2.vertices,
                              assume_unique=True)) == 0

def _join(def1, op, def2):
    return '(' + def1 + ' ' + op + ' ' + def2 + ')'

class Region(Struct):
    """
    Region defines a subset of a FE domain.

    Region kinds:

    - cell_only, facet_only, face_only, edge_only, vertex_only - only the
      specified entities are included, others are empty sets (so that the
      operators are still defined)
    - cell, facet, face, edge, vertex - entities of higher dimension are not
      included

    The 'cell' kind is the most general and it is the default.

    Region set-like operators: + (union), - (difference), * (intersection),
    followed by one of ('v', 'e', 'f', 'c', and 's') for vertices, edges,
    faces, cells, and facets.

    Created: 31.10.2005
    """
    __can = {
        'cell'        : (1, 1, 1, 1),
        'face'        : (1, 1, 1, 0),
        'edge'        : (1, 1, 0, 0),
        'vertex'      : (1, 0, 0, 0),
        'cell_only'   : (0, 0, 0, 1),
        'face_only'   : (0, 0, 1, 0),
        'edge_only'   : (0, 1, 0, 0),
        'vertex_only' : (1, 0, 0, 0),
    }

    __facet_kinds = {
        1 : {'facet' : 'vertex', 'facet_only' : 'vertex_only'},
        2 : {'facet' : 'edge',   'facet_only' : 'edge_only'},
        3 : {'facet' : 'face',   'facet_only' : 'face_only'},
    }

    __op_to_fun = {
        '+' : nm.union1d,
        '-' : nm.setdiff1d,
        '*' : nm.intersect1d,
    }

    @staticmethod
    def from_vertices(vertices, domain, name='region', kind='cell'):
        """
        Create a new region containing given vertices.

        Parameters
        ----------
        vertices : array
            The array of vertices.
        domain : Domain instance
            The domain containing the vertices.
        name : str, optional
            The name of the region.
        kind : str, optional
            The kind of the region.

        Returns
        -------
        obj : Region instance
            The new region.
        """
        obj = Region(name, 'given vertices', domain, '', kind=kind)
        obj.vertices = vertices

        return obj

    @staticmethod
    def from_facets(facets, domain, name='region', kind='facet', parent=None):
        """
        Create a new region containing given facets.

        Parameters
        ----------
        facets : array
            The array with indices to unique facets.
        domain : Domain instance
            The domain containing the facets.
        name : str, optional
            The name of the region.
        kind : str, optional
            The kind of the region.
        parent : str, optional
            The name of the parent region.

        Returns
        -------
        obj : Region instance
            The new region.
        """
        obj = Region(name, 'given faces', domain, '', kind=kind, parent=parent)
        obj.facets = facets

        return obj

    @staticmethod
    def from_cells(cells, domain, name='region', kind='cell', parent=None):
        """
        Create a new region containing given cells.

        Parameters
        ----------
        cells : array
            The array of cells.
        domain : Domain instance
            The domain containing the facets.
        name : str, optional
            The name of the region.
        kind : str, optional
            The kind of the region.
        parent : str, optional
            The name of the parent region.

        Returns
        -------
        obj : Region instance
            The new region.
        """
        obj = Region(name, 'given cells', domain, '', kind=kind, parent=parent)
        obj.cells = cells

        return obj

    def __init__(self, name, definition, domain, parse_def, kind='cell',
                 parent=None):
        """
        Create region instance.

        Parameters
        ----------
        name : str
            The region name, either given, or automatic for intermediate
            regions.
        definition : str
            The region selector definition.
        domain : Domain instance
            The domain of the region.
        parse_def : str
            The parsed definition of the region.
        kind : str
            The region kind - one of 'cell', 'facet', 'face', 'edge', 'vertex',
            'cell_only', ..., 'vertex_only'.
        parent : str, optional
            The name of the parent region.
        """
        tdim = domain.shape.tdim
        Struct.__init__(self,
                        name=name, definition=definition,
                        domain=domain, parse_def=parse_def,
                        n_v_max=domain.shape.n_nod, dim=domain.shape.dim,
                        tdim=tdim, kind_tdim=None,
                        entities=[None] * (tdim + 1),
                        kind=None, parent=parent, shape=None,
                        mirror_regions={}, mirror_maps={}, is_empty=False)
        self.set_kind(kind)

    def set_kind(self, kind):
        if kind == self.kind: return

        self.kind = kind
        if 'facet' in kind:
            self.true_kind = self.__facet_kinds[self.tdim][kind]

        else:
            self.true_kind = kind

        can = [bool(ii) for ii in self.__can[self.true_kind]]

        self.can_vertices = can[0]

        if self.tdim == 1:
            self.can = (can[0], can[3])
            self.can_cells = can[1]

        elif self.tdim == 2:
            self.can = (can[0], can[1], can[3])
            self.can_edges = can[1]
            self.can_cells = can[2]

        else:
            self.can = can
            self.can_edges = can[1]
            self.can_faces = can[2]
            self.can_cells = can[3]

        for ii, ican in enumerate(self.can):
            if not ican:
                self.entities[ii] = nm.empty(0, dtype=nm.uint32)

        self.set_kind_tdim()

    def set_kind_tdim(self):
        if 'vertex' in self.true_kind:
            self.kind_tdim = 0

        elif 'edge' in self.true_kind:
            self.kind_tdim = 1

        elif 'face' in self.true_kind:
            self.kind_tdim = 2

        elif 'cell' in self.true_kind:
            self.kind_tdim = self.tdim

    @property
    def vertices(self):
        if self.entities[0] is None:
            self._access(1)
            self.setup_from_highest(0)

        return self.entities[0]

    @vertices.setter
    def vertices(self, vals):
        if self.can_vertices:
            self.entities[0] = nm.asarray(vals, dtype=nm.uint32)

        else:
            raise ValueError('region "%s" cannot have vertices!' % self.name)

    @property
    def edges(self):
        if self.tdim <= 1:
            raise AttributeError('1D region has no edges!')

        if self.entities[1] is None:
            if 'edge' in self.true_kind:
                self.setup_from_vertices(1)

            else:
                self._access(2)
                self.setup_from_highest(1)

        return self.entities[1]

    @edges.setter
    def edges(self, vals):
        if self.can_edges:
            self.entities[1] = nm.asarray(vals, dtype=nm.uint32)

        else:
            raise ValueError('region "%s" cannot have edges!' % self.name)

    @property
    def faces(self):
        if self.tdim <= 2:
            raise AttributeError('1D or 2D region has no faces!')

        if self.entities[2] is None:
            if 'face' in self.true_kind:
                self.setup_from_vertices(2)

            else:
                self._access(3)
                self.setup_from_highest(2)

        return self.entities[2]

    @faces.setter
    def faces(self, vals):
        if self.can_faces:
            self.entities[2] = nm.asarray(vals, dtype=nm.uint32)

        else:
            raise ValueError('region "%s" cannot have faces!' % self.name)

    @property
    def facets(self):
        if self.tdim == 3:
            return self.faces

        elif self.tdim == 2:
            return self.edges

        else:
            return self.vertices

    @facets.setter
    def facets(self, vals):
        if self.tdim == 3:
            self.faces = vals

        elif self.tdim == 2:
            self.edges = vals

        else:
            self.vertices = vals

    @property
    def cells(self):
        if self.entities[self.tdim] is None:
            self.setup_from_vertices(self.tdim)
        return self.entities[self.tdim]

    @cells.setter
    def cells(self, vals):
        if self.can_cells:
            self.entities[self.tdim] = nm.asarray(vals, dtype=nm.uint32)

        else:
            raise ValueError('region "%s" cannot have cells!' % self.name)

    def _access(self, dim):
        """
        Helper to access region entities of dimension `dim`.
        """
        if dim == 0:
            self.vertices

        elif dim == 1:
            if self.tdim == 1:
                self.cells

            else:
                self.edges

        elif dim == 2:
            if self.tdim == 3:
                self.faces

            else:
                self.cells

        else:
            self.cells

    def setup_from_highest(self, dim, allow_lower=True, allow_empty=False):
        """
        Setup entities of topological dimension `dim` using the available
        entities of the highest topological dimension.
        """
        if not self.can[dim]: return

        for idim in range(self.tdim, -1, -1):
            if self.entities[idim] is not None:
                if self.entities[idim].shape[0] > 0:
                    break

        else:
            if not (allow_empty or self.is_empty):
                msg = 'region "%s" has no entities!'
                raise ValueError(msg % self.name)

            if self.entities[dim] is None:
                self.entities[dim] = nm.empty(0, dtype=nm.uint32)

            self.is_empty = True

            return

        cmesh = self.domain.cmesh
        if idim <= dim:
            if not (allow_lower or allow_empty):
                msg = 'setup_from_highest() can be used only with dim < %d'
                raise ValueError(msg % idim)

            if allow_lower:
                cmesh.setup_connectivity(dim, idim)
                ents = self.get_entities(idim)
                self.entities[dim] = cmesh.get_complete(dim, ents, idim)

            else:
                for idim in range(self.kind_tdim - 1, -1, -1):
                    self.entities[idim] = nm.empty(0, dtype=nm.uint32)

                self.is_empty = True

        else:
            cmesh.setup_connectivity(idim, dim)

            incident = cmesh.get_incident(dim, self.entities[idim], idim)
            self.entities[dim] = nm.unique(incident)

    def setup_from_vertices(self, dim):
        """
        Setup entities of topological dimension `dim` using the region
        vertices.
        """
        if not self.can[dim]: return

        cmesh = self.domain.cmesh
        cmesh.setup_connectivity(dim, 0)
        vv = self.vertices
        self.entities[dim] = cmesh.get_complete(dim, vv, 0)

    def finalize(self, allow_empty=False):
        """
        Initialize the entities corresponding to the region kind and regenerate
        all already existing (accessed) entities of lower topological dimension
        from the kind entities.
        """
        self._access(self.kind_tdim)
        is_empty = self.entities[self.kind_tdim].shape[0] == 0
        if allow_empty:
            self.is_empty = is_empty

        elif is_empty:
            raise ValueError('region "%s" has no entities!' % self.name)

        for idim in range(self.kind_tdim - 1, -1, -1):
            if self.can[idim] and self.entities[idim] is not None:
                try:
                    self.setup_from_highest(idim, allow_lower=False,
                                            allow_empty=allow_empty)

                except ValueError as exc:
                    msg = '\n'.join((str(exc),
                                     'fix region kind? (region: %s, kind: %s)'
                                     % (self.name, self.kind)))
                    raise ValueError(msg)

    def eval_op_vertices(self, other, op):
        parse_def = _join(self.parse_def, '%sv' % op, other.parse_def)
        tmp = self.light_copy('op', parse_def)
        tmp.vertices = self.__op_to_fun[op](self.vertices, other.vertices)

        return tmp

    def eval_op_edges(self, other, op):
        parse_def = _join(self.parse_def, '%se' % op, other.parse_def)
        tmp = self.light_copy('op', parse_def)
        tmp.edges = self.__op_to_fun[op](self.edges, other.edges)

        return tmp

    def eval_op_faces(self, other, op):
        parse_def = _join(self.parse_def, '%sf' % op, other.parse_def)
        tmp = self.light_copy('op', parse_def)
        tmp.faces = self.__op_to_fun[op](self.faces, other.faces)

        return tmp

    def eval_op_facets(self, other, op):
        parse_def = _join(self.parse_def, '%ss' % op, other.parse_def)
        tmp = self.light_copy('op', parse_def)
        tmp.facets = self.__op_to_fun[op](self.facets, other.facets)

        return tmp

    def eval_op_cells(self, other, op):
        parse_def = _join(self.parse_def, '%sc' % op, other.parse_def)
        tmp = self.light_copy('op', parse_def)
        tmp.cells = self.__op_to_fun[op](self.cells, other.cells)

        return tmp

    def light_copy(self, name, parse_def):
        return Region(name, self.definition, self.domain, parse_def,
                      kind=self.kind)

    def copy(self):
        """
        Vertices-based copy.
        """
        tmp = self.light_copy('copy', self.parse_def)
        tmp.vertices = copy(self.vertices)

        return tmp

    def delete_zero_faces(self, eps=1e-14):
        raise NotImplementedError

    def update_shape(self):
        """
        Update shape of each group according to region vertices, edges,
        faces and cells.
        """
        n_vertex = self.vertices.shape[0]
        n_cell = self.cells.shape[0]
        n_edge = self.edges.shape[0] if self.tdim > 1 else 0
        n_face = self.faces.shape[0] if self.tdim == 3 else 0
        n_facet = self.facets.shape[0]

        self.shape = Struct(n_vertex=n_vertex,
                            n_edge=n_edge,
                            n_face=n_face,
                            n_facet=n_facet,
                            n_cell=n_cell)

    def get_entities(self, dim):
        """
        Return mesh entities of dimension `dim`.
        """
        if dim <= self.tdim:
            self._access(dim)
            out = self.entities[dim]

        else:
            out = nm.empty(0, dtype=nm.uint32)

        return out

    def get_cells(self, true_cells_only=True):
        """
        Get cells of the region.

        Raises ValueError if `true_cells_only` is True and the region kind does
        not allow cells. For `true_cells_only` equal to False, cells incident
        to facets are returned if the region itself contains no cells. Obeys
        parent region, if given.
        """
        if self.cells.shape[0] == 0:
            if true_cells_only:
                msg = 'region %s has not true cells! (has kind: %s)' \
                      % (self.name, self.kind)
                raise ValueError(msg)

            else:
                # Has to be consistent with get_facet_indices()!
                cmesh = self.domain.cmesh
                cmesh.setup_connectivity(self.tdim - 1, self.tdim)
                out = cmesh.get_incident(self.tdim, self.facets, self.tdim - 1)

                if self.parent is not None:
                    pcells = self.domain.regions[self.parent].cells
                    ip = nm.in1d(out, pcells, assume_unique=False)
                    out = out[ip]

        else:
            out = self.cells

        return out

    def get_cell_indices(self, cells, true_cells_only=True):
        """
        Return indices of `cells` in the region cells.

        Raises ValueError if `true_cells_only` is True and the region kind does
        not allow cells. For `true_cells_only` equal to False, cells incident
        to facets are returned if the region itself contains no cells.

        Notes
        -----
        If the number of unique values in `cells` is smaller or equal to the
        number of cells in the region, all `cells` has to be also the region
        cells (`self` is a superset of `cells`). The region cells are
        considered depending on `true_cells_only`.

        Otherwise, indices of all cells in `self` that are in `cells` are
        returned.
        """
        fcells = self.get_cells(true_cells_only=true_cells_only)

        ucells = nm.unique(cells)
        ufcells = nm.unique(fcells)
        if not len(nm.intersect1d(ucells, ufcells, assume_unique=True)):
            return nm.array([], dtype=nm.int)

        if len(ucells) <= len(ufcells):
            # self is a superset of cells.
            ii = nm.searchsorted(fcells, cells)
            assert_((fcells[ii] == cells).all())

        else:
            aux = nm.searchsorted(cells, fcells)
            ii = nm.where(nm.take(cells, aux, mode='clip') == fcells)[0]

        return ii

    def get_facet_indices(self):
        """
        Return an array (per group) of (iel, ifa) for each facet. A facet can
        be in 1 (surface) or 2 (inner) cells.
        """
        cmesh = self.domain.cmesh
        cmesh.setup_connectivity(self.tdim - 1, self.tdim)

        facets = self.facets
        cells, offs = cmesh.get_incident(self.tdim, facets, self.tdim - 1,
                                         ret_offsets=True)

        if self.parent is not None:
            pcells = self.domain.regions[self.parent].cells
            ip = nm.in1d(cells, pcells, assume_unique=False)
            cells = cells[ip]

            counts = nm.diff(offs).astype(nm.int32)
            pos = nm.repeat(nm.arange(facets.shape[0], dtype=nm.int32), counts)
            new_counts = nm.bincount(pos, weights=ip).astype(nm.uint32)
            offs = nm.cumsum(nm.r_[0, new_counts], dtype=nm.uint32)

        ii = cmesh.get_local_ids(facets, self.tdim - 1, cells, offs, self.tdim)
        fis = nm.c_[cells, ii]

        return fis

    def setup_mirror_region(self, mirror_name=None, ret_name=False):
        """
        Find the corresponding mirror region, set up element mapping.
        """
        from sfepy.discrete.fem.mesh import find_map

        regions = self.domain.regions
        eopts = self.extra_options

        if (mirror_name is None) and (eopts is not None)\
            and ('mirror_region' in eopts):
            mirror_name = eopts['mirror_region']

        if mirror_name is not None:
            if mirror_name in self.mirror_regions:
                return

            mreg = regions[mirror_name]
            if self.vertices.shape[0] != mreg.vertices.shape[0]:
                raise ValueError('%s: incompatible mirror region! (%s)'
                    % (self.name, mreg.name))
            coors = self.domain.cmesh.coors
            coors1 = coors[self.vertices, :]
            coors2 = coors[mreg.vertices, :]
            shift = ((nm.sum(coors2, axis=0) - nm.sum(coors1, axis=0))
                / coors1.shape[0])
            vmap1, vmap2 = find_map(coors1, coors2 - shift, join=False)
            if vmap1.shape[0] != coors1.shape[0]:
                print(coors1[vmap1])
                print(coors2[vmap2])
                raise ValueError('cannot match vertices!')

            fconn = self.domain.cmesh.get_conn_as_graph(self.dim - 1, 0)
            cc = self.domain.cmesh.get_centroids(self.dim - 1)
            fmap1, fmap2 = find_map(cc[self.facets], cc[mreg.facets] - shift,
                                    join=False)
            if fmap1.shape[0] != self.facets.shape[0]:
                print(cc[self.facets][fmap1])
                print(cc[mreg.facets][fmap2])
                raise ValueError('cannot match facets!')

            mirror_map_i = nm.zeros_like(fmap1)
            mirror_map_i[fmap1] = fmap2
            mirror_map = nm.zeros_like(fmap1)
            mirror_map[fmap2] = fmap1
        else:
            for reg in regions:
                mirror_parent = regions.find(reg.parent)
                if mirror_parent is None: continue
                if ((reg is not self)
                    and nm.all(self.vertices == reg.vertices)):
                    mreg = reg
                    mirror_map = mirror_map_i = None
                    break
            else:
                raise ValueError('cannot find mirror region! (%s)' % self.name)

        self.mirror_regions[mreg.name] = mreg
        self.mirror_maps[mreg.name] = mirror_map
        mreg.mirror_regions[self.name] = self
        mreg.mirror_maps[self.name] = mirror_map_i

        if ret_name:
            return mreg.name

    def get_mirror_region(self, name):
        return self.mirror_regions[name]

    def get_n_cells(self, is_surface=False):
        """
        Get number of region cells.

        Parameters
        ----------
        is_surface : bool
            If True, number of edges or faces according to domain
            dimension is returned instead.

        Returns
        -------
        n_cells : int
            The number of cells.
        """
        if is_surface:
            return self.shape.n_facet

        else:
            return self.shape.n_cell

    def has_cells(self):
        return self.cells.size > 0

    def contains(self, other):
        """
        Return True in the region contains the `other` region.

        The check is performed using entities corresponding to the other region
        kind.
        """
        tdim = other.kind_tdim

        se = self.get_entities(tdim)
        oe = other.entities[tdim]

        return len(nm.intersect1d(se, oe))

    def get_charfun(self, by_cell=False, val_by_id=False):
        """
        Return the characteristic function of the region as a vector of values
        defined either in the mesh vertices (by_cell == False) or cells. The
        values are either 1 (val_by_id == False) or sequential id + 1.
        """
        if by_cell:
            chf = nm.zeros((self.domain.shape.n_el,), dtype=nm.float64)
            if val_by_id:
                chf[self.cells] = self.cells + 1
            else:
                chf[self.cells] = 1.0

        else:
            chf = nm.zeros((self.domain.shape.n_nod,), dtype=nm.float64)
            if val_by_id:
                chf[self.vertices] = self.vertices + 1
            else:
                chf[self.vertices] = 1.0

        return chf

    def get_edge_graph(self):
        """
        Return the graph of region edges as a sparse matrix having uid(k) + 1
        at (i, j) if vertex[i] is connected with vertex[j] by the edge k.

        Degenerate edges are ignored.
        """
        from scipy.sparse import csr_matrix

        cmesh = self.domain.cmesh

        e_verts = cmesh.get_incident(0, self.edges, 1)
        e_verts.shape = (e_verts.shape[0] // 2, 2)

        ii = nm.where(e_verts[:, 0] != e_verts[:, 1])[0]
        edges = self.edges[ii]
        e_verts = e_verts[ii]

        vals = edges + 1
        rows = e_verts[:, 0]
        cols = e_verts[:, 1]

        num = self.vertices.max() + 1
        graph = csr_matrix((vals, (rows, cols)), shape=(num, num))

        nnz = graph.nnz
        # Symmetrize.
        graph = graph + graph.T
        assert_(graph.nnz == 2 * nnz)

        return graph



"""Classes for probing values of Variables, for example, along a line,
using PyVTK library"""

from __future__ import absolute_import
import numpy as nm
import vtk
from vtk.util import numpy_support as vtknm
import os.path as osp

from sfepy.base.base import Struct, output
from sfepy.linalg import make_axis_rotation_matrix
from sfepy.postprocess.utils_vtk import get_vtk_from_mesh, get_vtk_from_file
from six.moves import range

vtk_version = vtk.vtkVersion().GetVTKMajorVersion()

class Probe(Struct):
    """
    Probe class.
    """

    def __init__(self, data, mesh, **kwargs):
        """
        Parameters
        ----------
        data : dict
            The output dictionary.
        mesh : Mesh
            The mesh.
        """

        Struct.__init__(self, name=mesh.name, **kwargs)

        self.mesh_name = mesh.name[mesh.name.rfind(osp.sep) + 1:]
        self.dim = mesh.dim
        self.vtkdata = get_vtk_from_mesh(mesh, data, 'probe_')

        self.vtkprobe = vtk.vtkProbeFilter()
        if vtk_version < 6:
            self.vtkprobe.SetSource(self.vtkdata)
        else:
            self.vtkprobe.SetSourceData(self.vtkdata)

        self.probes = {}
        self.probes_png = {}

    def new_vtk_polyline(self, points, closed=False):
        """
        Create the VTKPolyData object and store the line data.

        Parameters
        ----------
        points : array
            The line points.

        Returns
        -------
        vtkpd : VTK object
            VTKPolyData with the polyline.
        """

        npts = points.shape[0]
        pts = vtk.vtkPoints()
        pts.SetNumberOfPoints(npts)
        for ii in range(npts):
            pts.SetPoint(ii, points[ii,:])

        nlns = npts
        if closed:
            nlns += 1
        lns = vtk.vtkCellArray()
        lns.InsertNextCell(nlns)
        for ii in range(npts):
            lns.InsertCellPoint(ii)

        if closed:
            lns.InsertCellPoint(0)

        vtkpd = vtk.vtkPolyData()
        vtkpd.SetPoints(pts)
        vtkpd.SetLines(lns)
        if vtk_version < 6:
            vtkpd.Update()

        return vtkpd

    def add_line_probe(self, name, p0, p1, n_point):
        """
        Create the line probe - VTK object.

        Parameters
        ----------
        name : str
            The probe name.
        p0 : array_like
            The coordinates of the start point.
        p1 : array_like
            The coordinates of the end point.
        n_point : int
           The number of probe points.
        """

        line = vtk.vtkLineSource()
        line.SetPoint1(p0)
        line.SetPoint2(p1)
        line.SetResolution(n_point)
        line.Update()

        pars = nm.arange(n_point + 1) / nm.float(n_point)
        self.probes[name] = (line, pars)
        self.probes_png[name] = False

    def add_points_probe(self, name, coors):
        """
        Create the point probe - VTK object.

        Parameters
        ----------
        name : str
            The probe name.
        coors : array
            The coordinates of the probe points.
        """
        coors = nm.asarray(coors)
        npts = coors.shape[0]
        pts = vtk.vtkPoints()
        pts.SetNumberOfPoints(npts)
        for ii in range(npts):
            pts.SetPoint(ii, coors[ii, :])

        poly = vtk.vtkPolyData()
        poly.SetPoints(pts)
        if vtk_version < 6:
            vtk.Update()

        pars = nm.arange(npts + 1)
        self.probes[name] = (poly, pars)
        self.probes_png[name] = False

    def add_ray_probe(self, name, p0, dirvec, p_fun, n_point):
        """
        Create the ray (line) probe - VTK object.

        Parameters
        ----------
        name : str
            The probe name.
        p0 : array
            The coordinates of the start point.
        dirvec : array
            The probe direction vector.
        p_fun : function
            The function returning the probe parametrization along the dirvec
            direction.
        n_point : int
           The number of probe points.
        """

        p0 = nm.array(p0, dtype=nm.float64)
        dirvec = nm.array(dirvec, dtype=nm.float64)
        dirvec /= nm.linalg.norm(dirvec)

        pars = p_fun(nm.arange(n_point, dtype=nm.float64))
        points = p0 + dirvec * pars[:,None]

        ray = self.new_vtk_polyline(points)
        self.probes[name] = (ray, pars)
        self.probes_png[name] = False

    def add_circle_probe(self, name, centre, normal, radius, n_point):
        """
        Create the ray (line) probe - VTK object.

        Parameters
        ----------
        name : str
            The probe name.
        centre : array
            The coordinates of the circle center point.
        normal : array
             The normal vector perpendicular to the circle plane.
        radius : float
            The radius of the circle.
        n_point : int
           The number of probe points.
        """

        pars = nm.linspace(0.0, 2.0*nm.pi, n_point + 1)[:-1]

        # Create the points in xy plane, centered at the origin.
        x = radius * nm.cos(pars[:,None])
        y = radius * nm.sin(pars[:,None])

        if len(centre) == 3:
            z = nm.zeros((n_point, 1), dtype=nm.float64)
            points = nm.c_[x, y, z]

            # Rotate to satisfy the normal, shift to the centre.
            n1 = nm.array([0.0, 0.0, 1.0], dtype=nm.float64)
            axis = nm.cross(n1, normal)
            angle = nm.arccos(nm.dot(n1, normal))

            if nm.linalg.norm(axis) < 0.1:
                # n1 == self.normal
                rot_mtx = nm.eye(3, dtype=nm.float64)
            else:
                rot_mtx = make_axis_rotation_matrix(axis, angle)

            points = nm.dot(points, rot_mtx)

        else:
            points = nm.c_[x, y]

        points += centre

        circle = self.new_vtk_polyline(points, closed=True)
        self.probes[name] = (circle, pars)
        self.probes_png[name] = False

    def gen_mesh_probe_png(self, probe, png_filename):
        """
        Generate PNG image of the FE mesh.

        Parameters
        ----------
        probe : VTK objectstr
            The probe, VTKPolyData or VTKSource.
        png_filename : str
            The name of the output PNG file.
        """

        surface = vtk.vtkDataSetSurfaceFilter()
        if vtk_version < 6:
            surface.SetInput(self.vtkdata)

        else:
            surface.SetInputData(self.vtkdata)

        surface.Update()

        gf = vtk.vtkGraphicsFactory()
        gf.SetOffScreenOnlyMode(1)
        gf.SetUseMesaClasses(1)

        ifa = vtk.vtkImagingFactory()
        ifa.SetUseMesaClasses(1)

        mapper = vtk.vtkPolyDataMapper()
        if vtk_version < 6:
            mapper.SetInput(surface.GetOutput())

        else:
            mapper.SetInputData(surface.GetOutput())

        mapper.SetScalarModeToUseCellData()

        actor = vtk.vtkActor()
        actor.SetMapper(mapper)
        actor.GetProperty().SetOpacity(0.33)

        mapper2 = vtk.vtkPolyDataMapper()
        if hasattr(probe, 'GetOutput'):
            probe0 = probe.GetOutput()

        else:
            probe0 = probe

        if vtk_version < 6:
            mapper2.SetInput(probe0)

        else:
            mapper2.SetInputData(probe0)

        actor2 = vtk.vtkActor()
        actor2.SetMapper(mapper2)
        actor2.GetProperty().SetColor(0,0,0)
        actor2.GetProperty().SetLineWidth(2)
        ren = vtk.vtkRenderer()
        renWin = vtk.vtkRenderWindow()
        renWin.SetOffScreenRendering(1)
        renWin.AddRenderer(ren)
        ren.AddActor(actor)
        ren.AddActor(actor2)
        ren.SetBackground(1, 1, 1)
        renWin.Render()

        image = vtk.vtkWindowToImageFilter()
        image.SetInput(renWin)
        image.Update()

        writer = vtk.vtkPNGWriter()
        writer.SetFileName(png_filename)
        if vtk_version < 6:
            writer.SetInput(image.GetOutput())

        else:
            writer.SetInputData(image.GetOutput())

        writer.Write()

    def __call__(self, probe_name, variable, probe_view=False,
                 ret_points=False):
        """
        Do the probe for the given variable.

        Parameters
        ----------
        probe_name : str
            The name of previously defined probe.
        variable : str
            The variable to be probed.
        probe_view: bool
            If True, save the probe visualization.
        ret_points : bool
            If True, return also the probe points.

        Returns
        -------
        params : array
            The parametrization of the probe points.
        points : array, optional
            If `ret_points` is True, the coordinates of points corresponding to
            `params`, where the `variable` is evaluated.
        values : array
            The probe values in the points.
        """

        inp = self.probes[probe_name][0]
        params = self.probes[probe_name][1]

        if hasattr(inp, 'GetOutputPort'):
            self.vtkprobe.SetInputConnection(inp.GetOutputPort())

        else:
            if vtk_version < 6:
                self.vtkprobe.SetInput(inp)

            else:
                self.vtkprobe.SetInputData(inp)

        self.vtkprobe.Update()
        pdata = self.vtkprobe.GetOutput()
        values = vtknm.vtk_to_numpy(pdata.GetPointData().GetArray(variable))

        output('probing data')
        output('  probe name: %s' % probe_name)
        output('  variable: %s' % variable)
        output('  points: %s' % params.shape[0])

        if probe_view and not(self.probes_png[probe_name]):
            pngname = 'probe_%s_%s.png' % (self.mesh_name, probe_name)
            self.gen_mesh_probe_png(inp, pngname)
            self.probes_png[probe_name] = True

        if ret_points:
            points = vtknm.vtk_to_numpy(pdata.GetPoints().GetData())
            points = points[:, :self.dim]

            return params, points, values

        else:
            return params, values


class ProbeFromFile(Probe):
    """
    Probe class - read a given VTK file.
    """

    def __init__(self, filename, **kwargs):
        """
        Parameters
        ----------
        filename : dict
            The name of a VTK file.
        """

        bname = osp.splitext(osp.basename(filename))[0]
        Struct.__init__(self, name=bname, **kwargs)

        self.vtkdata = get_vtk_from_file(filename)
        self.mesh_name = bname
        self.dim = 3

        self.vtkprobe = vtk.vtkProbeFilter()
        if vtk_version < 6:
            self.vtkprobe.SetSource(self.vtkdata)
        else:
            self.vtkprobe.SetSourceData(self.vtkdata)

        self.probes = {}
        self.probes_png = {}



from __future__ import absolute_import
from sfepy.base.testing import TestCommon

class Test(TestCommon):

    @staticmethod
    def from_conf(conf, options):
        return Test(conf=conf, options=options)

    def test_tensors(self):
        import numpy as nm
        from sfepy.linalg import dot_sequences, insert_strided_axis

        ok = True

        a = nm.arange(1, 10).reshape(3, 3)
        b = nm.arange(9, 0, -1).reshape(3, 3)

        dab = nm.dot(a, b)
        dabt = nm.dot(a, b.T)
        datb = nm.dot(a.T, b)
        datbt = nm.dot(a.T, b.T)

        sa = insert_strided_axis(a, 0, 10)
        sb = insert_strided_axis(b, 0, 10)

        dsab = dot_sequences(sa, sb, mode='AB')
        _ok = nm.allclose(dab[None, ...], dsab, rtol=0.0, atol=1e-14)
        self.report('dot_sequences AB: %s' % _ok)
        ok = ok and _ok

        dsabt = dot_sequences(sa, sb, mode='ABT')
        _ok = nm.allclose(dabt[None, ...], dsabt, rtol=0.0, atol=1e-14)
        self.report('dot_sequences ABT: %s' % _ok)
        ok = ok and _ok

        dsatb = dot_sequences(sa, sb, mode='ATB')
        _ok = nm.allclose(datb[None, ...], dsatb, rtol=0.0, atol=1e-14)
        self.report('dot_sequences ATB: %s' % _ok)
        ok = ok and _ok

        dsatbt = dot_sequences(sa, sb, mode='ATBT')
        _ok = nm.allclose(datbt[None, ...], dsatbt, rtol=0.0, atol=1e-14)
        self.report('dot_sequences ATBT: %s' % _ok)
        ok = ok and _ok

        return ok

    def test_unique_rows(self):
        import numpy as nm
        from sfepy.linalg import unique_rows

        a = nm.arange(1, 10).reshape(3, 3)

        b = nm.r_[a, a]
        c = unique_rows(b)

        ok = (a == c).all()

        return ok

    def test_assemble1d(self):
        import numpy as nm
        from sfepy.linalg import assemble1d

        a = nm.arange(5)
        b = nm.arange(2)

        assemble1d(b, [1, 1, 1, 1, 0, 0], a[[0, 2, 3, 4, 1, 1]])

        ok = (b == [2, 10]).all()

        return ok

    def test_geometry(self):
        import numpy as nm
        from sfepy.linalg import get_face_areas

        a1 = get_face_areas([[0, 1, 2, 3]],
                            [[0, 0], [1, 0], [1, 1], [0, 1]])

        a2 = get_face_areas([[0, 1, 2, 3]],
                            [[0, 0, 2], [1, 0, 2], [1, 1, 2], [0, 1, 2]])
        ok = nm.allclose([a1, a2], [1, 1], rtol=0, atol=1e-15)

        return ok



# 10.07.2007, c
# last revision: 25.03.2008
from __future__ import absolute_import
from sfepy import data_dir
from sfepy.mechanics.matcoefs import stiffness_from_lame

filename_meshes = ['/meshes/3d/cube_medium_tetra.mesh',
                   '/meshes/3d/cube_medium_tetra.mesh',
                   '/meshes/3d/cube_medium_hexa.mesh']
filename_meshes = [data_dir + name for name in filename_meshes]

all_your_bases = [1, 2, 1]

filename_mesh = None

field_1 = {
    'name' : '3_displacement',
    'dtype' : 'real',
    'shape' : (3,),
    'region' : 'Omega',
    'approx_order' : None,
}

def get_pars( dim, full = False ):
    import numpy as nm
    sym = (dim + 1) * dim // 2
    lam = 1e1
    mu = 1e0
    o = nm.array( [1.] * dim + [0.] * (sym - dim), dtype = nm.float64 )
    oot = nm.outer( o, o )
    if full:
        return lam * oot + mu * nm.diag( o + 1.0 )
    else:
        return lam, mu

material_1 = {
    'name' : 'solid',
    'values' : {
        'Dijkl' : get_pars( 3, True ),
        'D' : stiffness_from_lame(3, get_pars(3)[0], get_pars(3)[1]),
        'lam' : get_pars(3)[0],
        'mu' : get_pars(3)[1],
    }
}

material_2 = {
    'name' : 'spring',
    'values' : {
        '.stiffness' : 1e0,
    }
}

variable_1 = {
    'name' : 'u',
    'kind' : 'unknown field',
    'field' : '3_displacement',
    'order' : 0,
}
variable_2 = {
    'name' : 'v',
    'kind' : 'test field',
    'field' : '3_displacement',
    'dual' : 'u',
}

region_1000 = {
    'name' : 'Omega',
    'select' : 'all',
}

region_1 = {
    'name' : 'Bottom',
    'select' : 'vertices in (z < -0.499)',
    'kind' : 'facet',
}
region_2 = {
    'name' : 'Top',
    'select' : 'vertices in (z > 0.499)',
    'kind' : 'facet',
}

ebc_1 = {
    'name' : 'Load',
    'region' : 'Top',
    'dofs' : {'u.2' : 0.1},
}

integral_1 = {
    'name' : 'i',
    'order' : 2,
}

equations_getpars = {
    'balance_of_forces' :
    """dw_lin_elastic.i.Omega(solid.Dijkl, v, u)
     = dw_point_lspring.i.Bottom(spring.stiffness, v, u)""",
}

equations_matcoefs = {
    'balance_of_forces' :
    """dw_lin_elastic.i.Omega(solid.D, v, u)
     = dw_point_lspring.i.Bottom(spring.stiffness, v, u)""",
}

equations_iso = {
    'balance_of_forces' :
    """dw_lin_elastic_iso.i.Omega(solid.lam, solid.mu, v, u)
     = dw_point_lspring.i.Bottom(spring.stiffness, v, u)""",
}

solver_0 = {
    'name' : 'ls',
    'kind' : 'ls.scipy_direct',
}

solver_1 = {
    'name' : 'newton',
    'kind' : 'nls.newton',
    'i_max'      : 1,
    'eps_a'      : 1e-10,
}

from sfepy.base.testing import TestCommon

##
# 10.07.2007, c
class Test( TestCommon ):
    tests = ['test_get_solution', 'test_linear_terms']

    ##
    # 10.07.2007, c
    def from_conf( conf, options ):
        return Test( conf = conf, options = options )
    from_conf = staticmethod( from_conf )

    ##
    # c: 25.03.2008, r: 25.03.2008
    def test_linear_terms( self ):
        ok = True
        for sols in self.solutions:
            ok = ok and self.compare_vectors(sols[0], sols[1],
                                             label1 = 'getpars',
                                             label2 = 'matcoefs')
            ok = ok and self.compare_vectors(sols[0], sols[2],
                                             label1 = 'getpars',
                                             label2 = 'iso')
        return ok

    ##
    # c: 10.07.2007, r: 25.03.2008
    def test_get_solution( self ):
        from sfepy.applications import solve_pde
        from sfepy.base.base import IndexedStruct
        import os.path as op

        ok = True
        self.solutions = []
        for ii, approx_order in enumerate(all_your_bases):
            fname = filename_meshes[ii]

            self.conf.filename_mesh = fname
            fields = {'field_1' : {
                          'name' : '3_displacement',
                          'dtype' : 'real',
                          'shape' : (3,),
                          'region' : 'Omega',
                          'approx_order' : approx_order,
                    }
            }
            self.conf.edit('fields', fields)
            self.report('mesh: %s, base: %s' % (fname, approx_order))
            status = IndexedStruct()

            self.report('getpars')
            self.conf.equations = self.conf.equations_getpars
            problem, state1 = solve_pde(self.conf, status=status,
                                        save_results=False)
            converged = status.nls_status.condition == 0
            ok = ok and converged
            self.report('converged: %s' % converged)

            self.report('matcoefs')
            self.conf.equations = self.conf.equations_matcoefs
            problem, state2 = solve_pde(self.conf, status=status,
                                        save_results=False)
            converged = status.nls_status.condition == 0
            ok = ok and converged
            self.report('converged: %s' % converged)

            self.report('iso')
            self.conf.equations = self.conf.equations_iso
            problem, state3 = solve_pde(self.conf, status=status,
                                        save_results=False)
            converged = status.nls_status.condition == 0
            ok = ok and converged
            self.report('converged: %s' % converged)

            self.solutions.append((state1(), state2(), state3()))

            name = op.join(self.options.out_dir,
                           '_'.join(('test_elasticity_small_strain',
                                     op.splitext(op.basename(fname))[0],
                                     '%d' % approx_order))
                           + '.vtk')
            problem.save_state(name, state1)

        return ok



from __future__ import absolute_import
input_name = '../examples/linear_elasticity/material_nonlinearity.py'
output_name = 'test_material_nonlinearity.vtk'

from tests_basic import TestInput
class Test(TestInput):
    pass



"""
Quadratic eigenvalue problem solvers.
"""
from __future__ import absolute_import

import numpy as nm
import scipy.sparse as sps

from sfepy.base.base import output, get_default
from sfepy.base.timing import Timer
from sfepy.linalg.utils import max_diff_csr
from sfepy.solvers.solvers import QuadraticEVPSolver

def standard_call(call):
    """
    Decorator handling argument preparation and timing for quadratic
    eigensolvers.
    """
    def _standard_call(self, mtx_m, mtx_d, mtx_k, n_eigs=None,
                       eigenvectors=None, status=None, conf=None, **kwargs):
        timer = Timer(start=True)

        conf = get_default(conf, self.conf)
        mtx_m = get_default(mtx_m, self.mtx_m)
        mtx_d = get_default(mtx_d, self.mtx_d)
        mtx_k = get_default(mtx_k, self.mtx_k)
        n_eigs = get_default(n_eigs, self.n_eigs)
        eigenvectors = get_default(eigenvectors, self.eigenvectors)
        status = get_default(status, self.status)

        result = call(self, mtx_m, mtx_d, mtx_k,
                      n_eigs, eigenvectors, status, conf,
                      **kwargs)

        elapsed = timer.stop()
        if status is not None:
            status['time'] = elapsed

        return result

    return _standard_call

class LQuadraticEVPSolver(QuadraticEVPSolver):
    """
    Quadratic eigenvalue problem solver based on the problem linearization.

    (w^2 M + w D + K) x = 0.
    """
    name = 'eig.qevp'

    _parameters = [
        ('method', "{'companion', 'cholesky'}", 'companion', False,
         'The linearization method.'),
        ('solver', 'dict', {'kind': 'eig.scipy', 'method': 'eig'}, False,
         """The configuration of an eigenvalue solver for
            the linearized problem (A - w B) x = 0."""),
        ('mode', "{'normal', 'inverted'}", 'normal', False,
         'Solve either A - w B (normal), or B - 1/w A (inverted).'),
        ('debug', 'bool', False, False,
         'If True, print debugging information.'),
    ]

    @standard_call
    def __call__(self, mtx_m, mtx_d, mtx_k, n_eigs=None,
                 eigenvectors=None, status=None, conf=None):
        if conf.debug:
            ssym = status['matrix_info'] = {}
            ssym['|M - M^T|'] = max_diff_csr(mtx_m, mtx_m.T)
            ssym['|D - D^T|'] = max_diff_csr(mtx_d, mtx_d.T)
            ssym['|K - K^T|'] = max_diff_csr(mtx_k, mtx_k.T)
            ssym['|M - M^H|'] = max_diff_csr(mtx_m, mtx_m.H)
            ssym['|D - D^H|'] = max_diff_csr(mtx_d, mtx_d.H)
            ssym['|K - K^H|'] = max_diff_csr(mtx_k, mtx_k.H)

        if conf.method == 'companion':
            mtx_eye = -sps.eye(mtx_m.shape[0], dtype=mtx_m.dtype)

            mtx_a = sps.bmat([[mtx_d, mtx_k],
                              [mtx_eye, None]])
            mtx_b = sps.bmat([[-mtx_m, None],
                              [None, mtx_eye]])

        elif conf.method == 'cholesky':
            from sksparse.cholmod import cholesky

            factor = cholesky(mtx_m)
            perm = factor.P()
            ir = nm.arange(len(perm))
            mtx_p = sps.coo_matrix((nm.ones_like(perm), (ir, perm)))
            mtx_l = mtx_p.T * factor.L()

            if conf.debug:
                ssym['|S - LL^T|'] = max_diff_csr(mtx_m, mtx_l * mtx_l.T)

            mtx_eye = sps.eye(mtx_l.shape[0], dtype=nm.float64)

            mtx_a = sps.bmat([[-mtx_k, None],
                              [None, mtx_eye]])
            mtx_b = sps.bmat([[mtx_d, mtx_l],
                              [mtx_l.T, None]])

        else:
            raise ValueError('unknown method! (%s)' % conf.method)

        if conf.debug:
            ssym['|A - A^T|'] = max_diff_csr(mtx_a, mtx_a.T)
            ssym['|A - A^H|'] = max_diff_csr(mtx_a, mtx_a.H)
            ssym['|B - B^T|'] = max_diff_csr(mtx_b, mtx_b.T)
            ssym['|B - B^H|'] = max_diff_csr(mtx_b, mtx_b.H)

            for key, val in sorted(ssym.items()):
                output('{}: {}'.format(key, val))

        if conf.mode == 'normal':
            out = self.solver(mtx_a, mtx_b, n_eigs=n_eigs,
                              eigenvectors=eigenvectors, status=status)

            if eigenvectors:
                eigs, vecs = out
                out = (eigs, vecs[:mtx_m.shape[0], :])

                if conf.debug:
                    res = mtx_a.dot(vecs) - eigs * mtx_b.dot(vecs)
                    status['lin. error'] = nm.linalg.norm(res, nm.inf)

        else:
            out = self.solver(mtx_b, mtx_a, n_eigs=n_eigs,
                              eigenvectors=eigenvectors, status=status)

            if eigenvectors:
                eigs, vecs = out
                out = (1.0 / eigs, vecs[:mtx_m.shape[0], :])

                if conf.debug:
                    res = (1.0 / eigs) * mtx_b.dot(vecs) -  mtx_a.dot(vecs)
                    status['lin. error'] = nm.linalg.norm(res, nm.inf)

            else:
                out = 1.0 / out

        if conf.debug and eigenvectors:
            eigs, vecs = out
            res = ((eigs**2 * (mtx_m.dot(vecs)))
                   + (eigs * (mtx_d.dot(vecs)))
                   + (mtx_k.dot(vecs)))
            status['error'] = nm.linalg.norm(res, nm.inf)

        return out



# -*- Mode: Python -*-
"""
C Mesh data structures and functions.
"""
cimport numpy as np

from libc.stdio cimport FILE, stdout

from types cimport uint32, int32, float64, complex128

cdef extern from 'string.h':
    void *memcpy(void *dest, void *src, size_t n)

cdef extern from 'common.h':
    void *pyalloc(size_t size)
    void pyfree(void *pp)
    void mem_statistics(int lineNo, char *funName,
                        char *fileName, char *dirName)
    size_t mem_get_cur_usage()
    size_t mem_get_max_usage()
    size_t mem_get_n_frags()

cdef extern from 'mesh.h':
    ctypedef struct Indices:
        uint32 *indices
        uint32 num

    ctypedef struct Mask:
        char *mask
        uint32 num
        uint32 n_true

    ctypedef struct MeshGeometry:
        uint32 num
        uint32 dim
        float64 *coors

    ctypedef struct MeshTopology:
        uint32 max_dim
        uint32 num[4]
        uint32 *cell_types
        uint32 *face_oris
        uint32 *edge_oris
        MeshConnectivity *conn[16]

    ctypedef struct MeshConnectivity:
        uint32 num
        uint32 n_incident
        uint32 *indices
        uint32 *offsets
        uint32 offset

    ctypedef struct LocalEntities:
        uint32 num
        MeshConnectivity **edges
        MeshConnectivity **faces

    ctypedef struct Mesh:
        MeshGeometry geometry[1]
        MeshTopology topology[1]
        LocalEntities entities[1]

    cdef int32 mesh_init(Mesh *mesh)
    cdef int32 mesh_free(Mesh *mesh)
    cdef int32 mesh_print(Mesh *mesh, FILE *file, int32 header_only)

    cdef int32 conn_alloc(MeshConnectivity *conn,
                          uint32 num, uint32 n_incident)
    cdef int32 conn_free(MeshConnectivity *conn)
    cdef int32 conn_print(MeshConnectivity *conn, FILE *file)

    cdef int32 mesh_set_coors(Mesh *mesh, float64 *coors, int32 num, int32 dim,
                              int32 tdim)

    cdef int32 mesh_setup_connectivity(Mesh *mesh, int32 d1, int32 d2)
    cdef int32 mesh_free_connectivity(Mesh *mesh, int32 d1, int32 d2)

    cdef uint32 mesh_count_incident(Mesh *mesh, int32 dim,
                                    Indices *entities, int32 dent)
    cdef int32 mesh_get_incident(Mesh *mesh,
                                 MeshConnectivity *incident, int32 dim,
                                 Indices *entities, int32 dent)
    cdef int32 mesh_get_local_ids(Mesh *mesh, Indices *local_ids,
                                  Indices *entities, int32 dent,
                                  MeshConnectivity *incident, int32 dim)
    cdef int32 mesh_select_complete(Mesh *mesh, Mask *mask, int32 dim,
                                    Indices *entities, int32 dent)
    cdef int32 mesh_get_centroids(Mesh *mesh, float64 *ccoors, int32 dim)
    cdef int32 mesh_get_volumes(Mesh *mesh, float64 *volumes, int32 dim)
    cdef int32 mesh_get_facet_normals(Mesh *mesh, float64 *normals,
                                      int32 which)

cdef class CConnectivity:
    """
    Notes
    -----

    The memory is allocated/freed in C - this class just wraps NumPy arrays
    around that data without copying.
    """
    cdef MeshConnectivity *conn

    cdef public np.ndarray indices
    cdef public np.ndarray offsets
    cdef public int num, n_incident, offset

    cdef _set_conn(self, MeshConnectivity *conn)

cdef class CMesh:
    cdef Mesh mesh[1]

    cdef readonly np.ndarray coors
    cdef readonly np.ndarray vertex_groups
    cdef readonly np.ndarray cell_types
    cdef readonly np.ndarray cell_groups # ig for each cell.
    cdef readonly list conns
    cdef readonly dict entities
    cdef readonly int n_coor, dim, n_el, tdim
    cdef readonly np.ndarray num # Numbers of topological entities.
    cdef readonly np.ndarray face_oris # Allocated in C.
    cdef readonly np.ndarray edge_oris # Allocated in C.
    cdef readonly np.ndarray facet_oris # face_oris in 3D, edge_oris in 2D

    cdef readonly dict key_to_index



# 04.06.2007, c
# last revision: 25.02.2008
from __future__ import absolute_import
from sfepy import data_dir

filename_mesh = data_dir + '/meshes/various_formats/small3d.mesh'

material_1 = {
    'name' : 'coef',
    'values' : {'coef' : 1.0},
}

region_1000 = {
    'name' : 'Omega',
    'select' : 'all',
}
region_1 = {
    'name' : 'Left',
    'select' : 'vertices in (x < -0.499)',
    'kind' : 'facet',
}
region_2 = {
    'name' : 'Right',
    'select' : 'vertices in (x > 0.499)',
    'kind' : 'facet',
}
region_3 = {
    'name' : 'Near',
    'select' : 'vertices in (y < -0.499)',
    'kind' : 'facet',
}
region_4 = {
    'name' : 'Far',
    'select' : 'vertices in (y > 0.499)',
    'kind' : 'facet',
}
region_5 = {
    'name' : 'Bottom',
    'select' : 'vertices in (z < -0.499)',
    'kind' : 'facet',
}
region_6 = {
    'name' : 'Top',
    'select' : 'vertices in (z > 0.499)',
    'kind' : 'facet',
}

field_1 = {
    'name' : '3_displacement',
    'dtype' : 'real',
    'shape' : (3,),
    'region' : 'Omega',
    'approx_order' : 2,
}

field_2 = {
    'name' : 'pressure',
    'dtype' : 'real',
    'shape' : (1,),
    'region' : 'Omega',
    'approx_order' : 1,
}

variable_1 = {
    'name' : 'u',
    'kind' : 'unknown field',
    'field' : '3_displacement',
    'order' : 0,
}
variable_2 = {
    'name' : 'v',
    'kind' : 'test field',
    'field' : '3_displacement',
    'dual' : 'u',
}
variable_3 = {
    'name' : 'p',
    'kind' : 'unknown field',
    'field' : 'pressure',
    'order' : 1,
}
variable_4 = {
    'name' : 'q',
    'kind' : 'test field',
    'field' : 'pressure',
    'dual' : 'p',
}

ebcs = {}
epbc_10 = {
    'name' : 'rl',
    'region' : ['Left', 'Right'],
    'dofs' : {'u.all' : 'u.all', 'p.0' : 'p.0'},
    'match' : 'match_x_plane',
}
epbc_12 = {
    'name' : 'tb',
    'region' : ['Top', 'Bottom'],
    'dofs' : {'u.all' : 'u.all', 'p.0' : 'p.0'},
    'match' : 'match_z_plane',
}
epbc_13 = {
    'name' : 'nf',
    'region' : ['Near', 'Far'],
    'dofs' : {'u.all' : 'u.all', 'p.0' : 'p.0'},
    'match' : 'match_y_plane',
}

from sfepy.discrete.fem.periodic import match_x_plane, match_y_plane, match_z_plane

functions = {
    'match_x_plane' : (match_x_plane,),
    'match_y_plane' : (match_y_plane,),
    'match_z_plane' : (match_z_plane,),
}

from test_periodic_bc_2d import Test



from __future__ import absolute_import

from sfepy.base.base import (Struct, Container, OneTypeList, assert_,
                             output, get_default, basestr)
from sfepy.base.timing import Timer
from .functions import ConstantFunction, ConstantFunctionByRegion
import six


class Materials(Container):

    @staticmethod
    def from_conf(conf, functions, wanted=None):
        """
        Construct Materials instance from configuration.
        """
        if wanted is None:
            wanted = list(conf.keys())

        objs = OneTypeList(Material)
        for key, mc in six.iteritems(conf):
            if key not in wanted: continue

            mat = Material.from_conf(mc, functions)
            objs.append(mat)

        obj = Materials(objs)
        return obj

    def reset(self):
        """Clear material data so that next materials.time_update() is
        performed even for stationary materials."""
        for mat in self:
            mat.reset()

    def time_update(self, ts, equations, mode='normal', problem=None,
                    verbose=True):
        """
        Update material parameters for given time, problem, and equations.

        Parameters
        ----------
        ts : TimeStepper instance
            The time stepper.
        equations : Equations instance
            The equations using the materials.
        mode : 'normal', 'update' or 'force'
            The update mode, see :func:`Material.time_update()`.
        problem : Problem instance, optional
            The problem that can be passed to user functions as a context.
        verbose : bool
            If False, reduce verbosity.
        """
        if verbose: output('updating materials...')
        timer = Timer(start=True)
        for mat in self:
            if verbose: output(' ', mat.name)
            mat.time_update(ts, equations, mode=mode, problem=problem)
        if verbose: output('...done in %.2f s' % timer.stop())

class Material(Struct):
    """
    A class holding constitutive and other material parameters.

    Example input::

        material_2 = {
           'name' : 'm',
           'values' : {'E' : 1.0},
        }

    Material parameters are passed to terms using the dot notation,
    i.e. 'm.E' in our example case.
    """
    @staticmethod
    def from_conf(conf, functions):
        """
        Construct Material instance from configuration.
        """
        kind = conf.get('kind', 'time-dependent')
        flags = conf.get('flags', {})

        function = conf.get('function', None)
        values = conf.get('values', None)

        if isinstance(function, basestr):
            function = functions[function]

        obj = Material(conf.name, kind, function, values, flags)

        return obj

    def __init__(self, name, kind='time-dependent',
                 function=None, values=None, flags=None, **kwargs):
        """
        Parameters
        ----------
        name : str
            The name of the material.
        kind : 'time-dependent' or 'stationary'
            The kind of the material.
        function : function
            The function for setting up the material values.
        values : dict
            Constant material values.
        flags : dict, optional
            Special flags.
        **kwargs : keyword arguments, optional
            Constant material values passed by their names.
        """
        Struct.__init__(self, name=name, kind=kind, is_constant=False)

        if (function is not None) and ((values is not None) or len(kwargs)):
            msg = 'material can have function or values but not both! (%s)' \
                  % self.name
            raise ValueError(msg)

        self.flags = get_default(flags, {})

        if hasattr(function, '__call__'):
            self.function = function

        elif (values is not None) or len(kwargs): # => function is None
            if isinstance(values, dict):
                key0 = list(values.keys())[0]
                assert_(isinstance(key0, str))

            else:
                key0 = None

            if (key0 and (not key0.startswith('.'))
                and isinstance(values[key0], dict)):
                self.function = ConstantFunctionByRegion(values)
                self.is_constant = True

            else:
                all_values = {}
                if values is not None:
                    all_values.update(values)
                all_values.update(kwargs)

                self.function = ConstantFunction(all_values)
                self.is_constant = True

        else: # => both values and function are None
            raise ValueError('material %s: neither "function" nor "values"'
                             ' arguments (or keyword arguments) given!'
                             % self.name)

        self.reset()

    def iter_terms(self, equations, only_new=True):
        """
        Iterate terms for which the material data should be evaluated.
        """
        if equations is None: return

        for equation in equations:
            for term in equation.terms:
                names = [ii[0] for ii in term.names.material]
                if self.name not in names: continue

                key = term.get_qp_key()
                if only_new and (key in self.datas): continue

                self.datas.setdefault(key, {})

                yield key, term

    def set_data(self, key, qps, data):
        """
        Set the material data in quadrature points.

        Parameters
        ----------
        key : tuple
            The (region_name, integral_name) data key.
        qps : Struct
            Information about the quadrature points.
        data : dict
            The material data.
        """
        # Restore shape to (n_el, n_qp, ...) until the C
        # core is rewritten to work with a bunch of physical
        # point values only.
        new_data = {}
        if data is not None:
            for dkey, val in six.iteritems(data):
                if val.ndim != 3:
                    raise ValueError('material parameter array must have'
                                     " three dimensions! ('%s' has %d)"
                                     % (dkey, val.ndim))
                new_data[dkey] = val.reshape(qps.get_shape(val.shape))

        self.datas[key] = new_data

    def update_data(self, key, ts, equations, term, problem=None):
        """
        Update the material parameters in quadrature points.

        Parameters
        ----------
        key : tuple
            The (region_name, integral_name) data key.
        ts : TimeStepper
            The time stepper.
        equations : Equations
            The equations for which the update occurs.
        term : Term
            The term for which the update occurs.
        problem : Problem, optional
            The problem definition for which the update occurs.
        """
        self.datas.setdefault(key, {})

        qps = term.get_physical_qps()
        coors = qps.values
        data = self.function(ts, coors, mode='qp',
                             equations=equations, term=term, problem=problem,
                             **self.extra_args)

        self.set_data(key, qps, data)

    def update_special_data(self, ts, equations, problem=None):
        """
        Update the special material parameters.

        Parameters
        ----------
        ts : TimeStepper
            The time stepper.
        equations : Equations
            The equations for which the update occurs.
        problem : Problem, optional
            The problem definition for which the update occurs.
        """
        if 'special' in self.datas: return

        # Special function values (e.g. flags).
        datas = self.function(ts, None, mode='special',
                              problem=problem, equations=equations,
                              **self.extra_args)
        if datas is not None:
            self.datas['special'] = datas
            self.special_names.update(list(datas.keys()))

    def update_special_constant_data(self, equations=None, problem=None):
        """
        Update the special constant material parameters.

        Parameters
        ----------
        equations : Equations
            The equations for which the update occurs.
        problem : Problem, optional
            The problem definition for which the update occurs.
        """
        if 'special_constant' in self.datas: return
        if not self.flags.get('special_constant'): return

        # Special constant values.
        datas = self.function(None, None, mode='special_constant',
                              problem=problem, equations=equations)
        self.datas['special_constant'] = datas
        self.constant_names.update(list(datas.keys()))

    def time_update(self, ts, equations, mode='normal', problem=None):
        """
        Evaluate material parameters in physical quadrature points.

        Parameters
        ----------
        ts : TimeStepper instance
            The time stepper.
        equations : Equations instance
            The equations using the materials.
        mode : 'normal', 'update' or 'force'
            The update mode. In 'force' mode, ``self.datas`` is cleared and all
            updates are redone. In 'update' mode, existing data are preserved
            and new can be added. The 'normal' mode depends on other
            attributes: for stationary (``self.kind == 'stationary'``)
            materials and materials in 'user' mode, nothing is done if
            ``self.datas`` is not empty. For time-dependent materials
            (``self.kind == 'time-dependent'``, the default) that are not
            constant, i.e., are given by a user function, 'normal' mode behaves
            like 'force' mode. For constant materials it behaves like 'update'
            mode - existing data are reused.
        problem : Problem instance, optional
            The problem that can be passed to user functions as a context.
        """
        if mode == 'force':
            self.datas = {}

        elif self.datas:
            if mode == 'normal':
                if (self.mode == 'user') or (self.kind == 'stationary'):
                    return

                elif not self.is_constant:
                    self.datas = {}

        for key, term in self.iter_terms(equations):
            self.update_data(key, ts, equations, term, problem=problem)

        self.update_special_data(ts, equations, problem=problem)
        self.update_special_constant_data(equations, problem=problem)

    def get_keys(self, region_name=None):
        """
        Get all data keys.

        Parameters
        ----------
        region_name : str
            If not None, only keys with this region are returned.
        """
        if not self.datas:
            keys = None

        elif region_name is None:
            keys = list(self.datas.keys())

        else:
            keys = [key for key in self.datas.keys()
                    if (isinstance(key, tuple) and key[0] == region_name)]

        return keys

    def set_all_data(self, datas):
        """
        Use the provided data, set mode to 'user'.
        """
        self.mode = 'user'
        self.datas = datas

    def set_function(self, function):
        self.function = function
        self.reset()

    def reset(self):
        """
        Clear all data created by a call to ``time_update()``, set
        ``self.mode`` to ``None``.
        """
        self.mode = None
        self.datas = {}
        self.special_names = set()
        self.constant_names = set()
        self.extra_args = {}

    def set_extra_args(self, **extra_args):
        """Extra arguments passed tu the material function."""
        self.extra_args = extra_args

    def get_data(self, key, name):
        """`name` can be a dict - then a Struct instance with data as
        attributes named as the dict keys is returned."""

        if isinstance(name, basestr):
            return self._get_data(key, name)
        else:
            out = Struct()
            for key, item in six.iteritems(name):
                setattr(out, key, self._get_data(key, item))
            return out

    def _get_data(self, key, name):
        if name is None:
            msg = 'material arguments must use the dot notation!\n'\
                  '(material: %s, key: %s)' % (self.name, key)
            raise ValueError(msg)

        if not self.datas:
            raise ValueError('material data not set! (call time_update())')

        if name in self.special_names:
            # key ignored.
            return self.datas['special'][name]

        else:
            datas = self.datas[key]

            if isinstance(datas, Struct):
                return getattr(datas, name)

            elif datas:
                return datas[name]

    def get_constant_data(self, name):
        """Get constant data by name."""
        if name in self.constant_names:
            # no key.
            return self.datas['special_constant'][name]
        else:
            raise ValueError('material %s has no constant %s!'
                             % (self.name, name))

    def reduce_on_datas(self, reduce_fun, init=0.0):
        """For non-special values only!"""
        out = {}.fromkeys(list(self.datas[list(self.datas.keys())[0]].keys()), init)
        for data in six.itervalues(self.datas):
            for key, val in six.iteritems(data):
                out[key] = reduce_fun(out[key], val)

        return out



# -*- coding: utf-8
r"""
Homogenization of the Darcy flow in a thin porous layer.
The reference cell is composed of the matrix representing the dual porosity
and of two disconnected channels representing the primary porosity,
see paper [1].

[1] E. Rohan, V. Lukeš: Modeling Tissue Perfusion Using a Homogenized
Model with Layer-wise Decomposition. IFAC Proceedings Volumes 45(2), 2012,
pages 1029-1034.
https://doi.org/10.3182/20120215-3-AT-3016.00182
"""

from __future__ import absolute_import
from sfepy.discrete.fem.periodic import match_x_plane, match_y_plane
import sfepy.homogenization.coefs_base as cb
import numpy as nm
from sfepy import data_dir
import six
from six.moves import range

def get_mats(pk, ph, pe, dim):
    m1 = nm.eye(dim, dtype=nm.float64) * pk
    m1[-1, -1] = pk / ph
    m2 = nm.eye(dim, dtype=nm.float64) * pk
    m2[-1, -1] = pk / ph ** 2

    return m1, m2


def recovery_perf(pb, corrs, macro):
    from sfepy.homogenization.recovery import compute_p_from_macro
    from sfepy.base.base import Struct

    slev = ''

    micro_nnod = pb.domain.mesh.n_nod

    centre_Y = nm.sum(pb.domain.mesh.coors, axis=0) / micro_nnod
    nodes_Y = {}

    channels = {}
    for k in six.iterkeys(macro):
        if 'press' in k:
            channels[k[-1]] = 1

    channels = list(channels.keys())

    varnames = ['pM']
    for ch in channels:
        nodes_Y[ch] = pb.domain.regions['Y' + ch].vertices
        varnames.append('p' + ch)

    pvars = pb.create_variables(varnames)

    press = {}

    # matrix
    press['M'] = \
       corrs['corrs_%s_gamma_p' % pb_def['name']]['pM'] * macro['g_p'] + \
       corrs['corrs_%s_gamma_m' % pb_def['name']]['pM'] * macro['g_m']

    out = {}
    # channels
    for ch in channels:
        press_mac = macro['press' + ch][0, 0]
        press_mac_grad = macro['pressg' + ch]
        nnod = corrs['corrs_%s_pi%s' % (pb_def['name'], ch)]\
            ['p%s_0' % ch].shape[0]

        press_mic = nm.zeros((nnod, 1))
        for key, val in \
          six.iteritems(corrs['corrs_%s_pi%s' % (pb_def['name'], ch)]):
            kk = int(key[-1])
            press_mic += val * press_mac_grad[kk, 0]

        for key in six.iterkeys(corrs):
            if ('_gamma_' + ch in key):
                kk = int(key[-1]) - 1
                press_mic += corrs[key]['p' + ch] * macro['g' + ch][kk]

        press_mic += \
          compute_p_from_macro(press_mac_grad[nm.newaxis,nm.newaxis, :, :],
                               micro_coors[nodes_Y[ch]], 0,
                               centre=centre_Y, extdim=-1).reshape((nnod, 1))

        press[ch] = press_mac + eps0 * press_mic

        out[slev + 'p' + ch] = Struct(name='output_data',
                                      mode='vertex',
                                      data=press[ch],
                                      var_name='p' + ch,
                                      dofs=None)

        pvars['p' + ch].set_data(press_mic)
        dvel = pb.evaluate('ev_diffusion_velocity.iV.Y%s(mat1%s.k, p%s)'
                           % (ch, ch, ch),
                           var_dict={'p' + ch: pvars['p' + ch]},
                           mode='el_avg')

        out[slev + 'w' + ch] = Struct(name='output_data',
                                      mode='cell',
                                      data=dvel,
                                      var_name='w' + ch,
                                      dofs=None)

        press['M'] += corrs['corrs_%s_eta%s' % (pb_def['name'], ch)]['pM']\
            * press_mac

    pvars['pM'].set_data(press['M'])
    dvel = pb.evaluate('%e * ev_diffusion_velocity.iV.YM(mat1M.k, pM)' % eps0,
                       var_dict={'pM': pvars['pM']}, mode='el_avg')

    out[slev + 'pM'] = Struct(name='output_data',
                              mode='vertex',
                              dat=press['M'],
                              var_name='pM',
                              dofs=None)

    out[slev + 'wM'] = Struct(name='output_data',
                              mode='cell',
                              data=dvel,
                              var_name='wM',
                              dofs=None)

    return out


geoms = {
    '2_4': ['2_4_Q1', '2', 5],
    '3_8': ['3_8_Q1', '4', 5],
    '3_4': ['3_4_P1', '3', 3],
}

pb_def = {
    'name': '3d_2ch',
    'mesh_filename': data_dir + '/meshes/3d/perfusion_micro3d.mesh',
    'dim': 3,
    'geom': geoms['3_4'],
    'eps0': 1.0e-2,
    'param_h': 1.0,
    'param_kappa_m': 0.1,
    'matrix_mat_el_grp': 3,
    'channels': {
        'A': {
            'mat_el_grp': 1,
            'fix_nd_grp': (4, 1),
            'io_nd_grp': [1, 2, 3],
            'param_kappa_ch': 1.0,
        },
        'B': {
            'mat_el_grp': 2,
            'fix_nd_grp': (14, 11),
            'io_nd_grp': [11, 12, 13],
            'param_kappa_ch': 2.0,
        },
    },
}

filename_mesh = pb_def['mesh_filename']
eps0 = pb_def['eps0']
param_h = pb_def['param_h']

# integrals
integrals = {
    'iV': 2,
    'iS': 2,
}

functions = {
    'match_x_plane': (match_x_plane,),
    'match_y_plane': (match_y_plane,),
}

aux = []
for ch, val in six.iteritems(pb_def['channels']):
    aux.append('r.bYM' + ch)

# basic regions
regions = {
    'Y': 'all',
    'YM': 'cells of group %d' % pb_def['matrix_mat_el_grp'],
    # periodic boundaries
    'Pl': ('vertices in (x < 0.001)', 'facet'),
    'Pr': ('vertices in (x > 0.999)', 'facet'),
    'PlYM': ('r.Pl *v r.YM', 'facet'),
    'PrYM': ('r.Pr *v r.YM', 'facet'),
    'bYMp': ('r.bYp *v r.YM', 'facet', 'YM'),
    'bYMm': ('r.bYm *v r.YM', 'facet', 'YM'),
    'bYMpm': ('r.bYMp +v r.bYMm', 'facet', 'YM'),
}

# matrix/channel boundaries
regions.update({
    'bYMchs': (' +v '.join(aux), 'facet', 'YM'),
    'YMmchs': 'r.YM -v r.bYMchs',
})

# boundary conditions Gamma+/-
ebcs = {
    'gamma_pm_bYMchs': ('bYMchs', {'pM.0': 0.0}),
    'gamma_pm_YMmchs': ('YMmchs', {'pM.0': 1.0}),
}

# periodic boundary conditions - matrix, X-direction
epbcs = {'periodic_xYM': (['PlYM', 'PrYM'], {'pM.0': 'pM.0'}, 'match_x_plane')}
lcbcs = {}

all_periodicYM = ['periodic_%sYM' % ii for ii in ['x', 'y'][:pb_def['dim']-1]]
all_periodicY = {}

if pb_def['dim'] == 2:
    regions.update({
        'bYm': ('vertices in (y < 0.001)', 'facet'),
        'bYp':  ('vertices in (y > 0.999)', 'facet'),
    })
if pb_def['dim'] == 3:
    regions.update({
        'Pn': ('vertices in (y < 0.001)', 'facet'),
        'Pf': ('vertices in (y > 0.999)', 'facet'),
        'PnYM': ('r.Pn *v r.YM', 'facet'),
        'PfYM': ('r.Pf *v r.YM', 'facet'),
        'bYm': ('vertices in (z < 0.001)', 'facet'),
        'bYp':  ('vertices in (z > 0.999)', 'facet'),
    })
    # periodic boundary conditions - matrix, Y-direction
    epbcs.update({
        'periodic_yYM': (['PnYM', 'PfYM'], {'pM.0': 'pM.0'}, 'match_y_plane'),
    })

reg_io = {}
ebcs_eta = {}
ebcs_gamma = {}

# generate regions, ebcs, epbcs
for ch, val in six.iteritems(pb_def['channels']):

    all_periodicY[ch] = ['periodic_%sY%s' % (ii, ch)
                         for ii in ['x', 'y'][:pb_def['dim']-1]]

    # channels: YA, fixedYA, bYMA (matrix/channel boundaries)
    regions.update({
        'Y' + ch: 'cells of group %d' % val['mat_el_grp'],
        'bYM' + ch: ('r.YM *v r.Y' + ch, 'facet', 'YM'),
        'PlY' + ch: ('r.Pl *v r.Y' + ch, 'facet'),
        'PrY' + ch: ('r.Pr *v r.Y' + ch, 'facet'),
    })

    if 'fix_nd_grp' in val:
        regions.update({
            'fixedY' + ch: ('vertices of group %d' % val['fix_nd_grp'][0],
                            'vertex'),
        })

    ebcs_eta[ch] = []
    for ch2, val2 in six.iteritems(pb_def['channels']):
        aux = 'eta%s_bYM%s' % (ch, ch2)
        if ch2 == ch:
            ebcs.update({aux: ('bYM' + ch2, {'pM.0': 1.0})})
        else:
            ebcs.update({aux: ('bYM' + ch2, {'pM.0': 0.0})})

        ebcs_eta[ch].append(aux)

    # boundary conditions
    # periodic boundary conditions - channels, X-direction
    epbcs.update({
            'periodic_xY' + ch: (['PlY' + ch, 'PrY' + ch],
                                 {'p%s.0' % ch: 'p%s.0' % ch},
                                 'match_x_plane'),
    })

    if pb_def['dim'] == 3:
        regions.update({
                'PnY' + ch: ('r.Pn *v r.Y' + ch, 'facet'),
                'PfY' + ch: ('r.Pf *v r.Y' + ch, 'facet'),
        })
        # periodic boundary conditions - channels, Y-direction
        epbcs.update({
                'periodic_yY' + ch: (['PnY' + ch, 'PfY' + ch],
                                     {'p%s.0' % ch: 'p%s.0' % ch},
                                     'match_y_plane'),
        })

    reg_io[ch] = []
    aux_bY = []
    # channel: inputs/outputs
    for i_io in range(len(val['io_nd_grp'])):
        io = '%s_%d' % (ch, i_io+1)

        # regions
        aux = val['io_nd_grp'][i_io]
        if 'fix_nd_grp' in val and val['fix_nd_grp'][1] == aux:
            regions.update({
                'bY%s' % io: ('vertices of group %d +v r.fixedY%s' % (aux, ch),
                              'facet', 'Y%s' % ch),
            })
        else:
            regions.update({
                'bY%s' % io: ('vertices of group %d' % aux,
                              'facet', 'Y%s' % ch),
            })

        aux_bY.append('r.bY%s' % io)
        reg_io[ch].append('bY%s' % io)

    regions.update({
        'bY' + ch: (' +v '.join(aux_bY), 'facet', 'Y' + ch),
    })

    # channel: inputs/outputs
    for i_io in range(len(val['io_nd_grp'])):
        io = '%s_%d' % (ch, i_io + 1)
        ion = '%s_n%d' % (ch, i_io + 1)
        regions.update({
            'bY%s' % ion: ('r.bY%s -v r.bY%s' % (ch, io), 'facet', 'Y%s' % ch),
        })

        # boundary conditions
        aux = 'fix_p%s_bY%s' % (ch, ion)
        ebcs.update({
            aux: ('bY%s' % ion, {'p%s.0' % ch: 0.0}),
        })

    lcbcs.update({
        'imv' + ch: ('Y' + ch, {'ls%s.all' % ch: None}, None,
                     'integral_mean_value')
    })


matk1, matk2 = get_mats(pb_def['param_kappa_m'], param_h, eps0, pb_def['dim'])

materials = {
    'mat1M': ({'k': matk1},),
    'mat2M': ({'k': matk2},),
}

fields = {
    'corrector_M': ('real', 'scalar', 'YM', 1),
    'vel_M': ('real', 'vector', 'YM', 1),
    'vol_all': ('real', 'scalar', 'Y', 1),
}

variables = {
    'pM': ('unknown field', 'corrector_M'),
    'qM': ('test field', 'corrector_M', 'pM'),
    'Pi_M': ('parameter field', 'corrector_M', '(set-to-None)'),
    'corr_M': ('parameter field', 'corrector_M', '(set-to-None)'),
    'corr1_M': ('parameter field', 'corrector_M', '(set-to-None)'),
    'corr2_M': ('parameter field', 'corrector_M', '(set-to-None)'),
    'wM': ('parameter field', 'vel_M', '(set-to-None)'),
    'vol_all': ('parameter field', 'vol_all', '(set-to-None)'),
}

# generate regions for channel inputs/outputs
for ch, val in six.iteritems(pb_def['channels']):

    matk1, matk2 = get_mats(val['param_kappa_ch'],  param_h,
                            eps0, pb_def['dim'])
    materials.update({
        'mat1' + ch: ({'k': matk1},),
        'mat2' + ch: ({'k': matk2},),
    })

    fields.update({
        'corrector_' + ch: ('real', 'scalar', 'Y' + ch, 1),
        'vel_' + ch: ('real', 'vector', 'Y' + ch, 1),
    })

    variables.update({
        'p' + ch: ('unknown field', 'corrector_' + ch),
        'q' + ch: ('test field', 'corrector_' + ch, 'p' + ch),
        'Pi_' + ch: ('parameter field', 'corrector_' + ch, '(set-to-None)'),
        'corr1_' + ch: ('parameter field', 'corrector_' + ch, '(set-to-None)'),
        'corr2_' + ch: ('parameter field', 'corrector_' + ch, '(set-to-None)'),
        'w' + ch: ('unknown field', 'vel_' + ch),
        # lagrange mutltipliers - integral mean value
        'ls' + ch: ('unknown field', 'corrector_' + ch),
        'lv' + ch: ('test field', 'corrector_' + ch, 'ls' + ch),
    })

options = {
    'coefs': 'coefs',
    'requirements': 'requirements',
    'ls': 'ls',  # linear solver to use
    'volumes': {
        'total': {
            'variables': ['vol_all'],
            'expression': """d_volume.iV.Y(vol_all)""",
        },
        'one': {
            'value': 1.0,
        }
    },
    'output_dir': './output',
    'file_per_var': True,
    'coefs_filename': 'coefs_perf_' + pb_def['name'],
    'coefs_info': {'eps0': eps0},
    'recovery_hook': 'recovery_perf',
    'multiprocessing': False,
}

for ipm in ['p', 'm']:
    options['volumes'].update({
        'bYM' + ipm: {
            'variables': ['pM'],
            'expression': "d_surface.iS.bYM%s(pM)" % ipm,
        },
        'bY' + ipm: {
            'variables': ['vol_all'],
            'expression': "d_surface.iS.bY%s(vol_all)" % ipm,
        }
    })

for ch in six.iterkeys(reg_io):
    for ireg in reg_io[ch]:
        options['volumes'].update({
            ireg: {
                'variables': ['p' + ch],
                'expression': "d_surface.iS.%s(p%s)" % (ireg, ch),
            }
        })

coefs = {
    'vol_bYMpm': {
        'regions': ['bYMp', 'bYMm'],
        'expression': 'd_surface.iS.%s(pM)',
        'class': cb.VolumeFractions,
    },
    'filenames': {},
}

requirements = {
    'corrs_one_YM': {
        'variable': ['pM'],
        'ebcs': ['gamma_pm_YMmchs', 'gamma_pm_bYMchs'],
        'epbcs': [],
        'save_name': 'corrs_one_YM',
        'class': cb.CorrSetBCS,
    },
}

for ipm in ['p', 'm']:
    requirements.update({
        'corrs_gamma_' + ipm: {
            'requires': [],
            'ebcs': ['gamma_pm_bYMchs'],
            'epbcs': all_periodicYM,
            'equations': {
                'eq_gamma_pm': """dw_diffusion.iV.YM(mat2M.k, qM, pM) =
                             %e * dw_surface_integrate.iS.bYM%s(qM)"""
                               % (1.0/param_h, ipm),
                },
            'class': cb.CorrOne,
            'save_name': 'corrs_%s_gamma_%s' % (pb_def['name'], ipm),
        },
    })

    for ipm2 in ['p', 'm']:
        coefs.update({
            'H' + ipm + ipm2: {  # test+
                'requires': ['corrs_gamma_' + ipm],
                'set_variables': [('corr_M', 'corrs_gamma_' + ipm, 'pM')],
                'expression': 'ev_surface_integrate.iS.bYM%s(corr_M)' % ipm2,
                'set_volume': 'bYp',
                'class': cb.CoefOne,
            },
        })


def get_channel(keys, bn):
    for ii in keys:
        if bn in ii:
            return ii[(ii.rfind(bn) + len(bn)):]

    return None


def set_corrpis(variables, ir, ic, mode, **kwargs):
    ch = get_channel(list(kwargs.keys()), 'pis_')
    pis = kwargs['pis_' + ch]
    corrs_pi = kwargs['corrs_pi' + ch]

    if mode == 'row':
        val = pis.states[ir]['p' + ch] + corrs_pi.states[ir]['p' + ch]
        variables['corr1_' + ch].set_data(val)
    elif mode == 'col':
        val = pis.states[ic]['p' + ch] + corrs_pi.states[ic]['p' + ch]
        variables['corr2_' + ch].set_data(val)


def set_corr_S(variables, ir, *args, **kwargs):
    ch = get_channel(list(kwargs.keys()), 'pis_')
    io = get_channel(list(kwargs.keys()), 'corrs_gamma_')

    pis = kwargs['pis_' + ch]
    corrs_gamma = kwargs['corrs_gamma_' + io]

    pi = pis.states[ir]['p' + ch]
    val = corrs_gamma.state['p' + ch]
    variables['corr1_' + ch].set_data(pi)
    variables['corr2_' + ch].set_data(val)


def set_corr_cc(variables, ir, *args, **kwargs):
    ch = get_channel(list(kwargs.keys()), 'pis_')
    pis = kwargs['pis_' + ch]
    corrs_pi = kwargs['corrs_pi' + ch]

    pi = pis.states[ir]['p' + ch]
    pi = pi - nm.mean(pi)
    val = pi + corrs_pi.states[ir]['p' + ch]
    variables['corr1_' + ch].set_data(val)


for ch, val in six.iteritems(pb_def['channels']):
    coefs.update({
        'G' + ch: {  # test+
            'requires': ['corrs_one' + ch, 'corrs_eta' + ch],
            'set_variables': [('corr1_M', 'corrs_one' + ch, 'pM'),
                              ('corr2_M', 'corrs_eta' + ch, 'pM')],
            'expression': 'dw_diffusion.iV.YM(mat2M.k, corr1_M, corr2_M)',
            'class': cb.CoefOne,
        },
        'K' + ch: {  # test+
            'requires': ['pis_' + ch, 'corrs_pi' + ch],
            'set_variables': set_corrpis,
            'expression': 'dw_diffusion.iV.Y%s(mat2%s.k, corr1_%s, corr2_%s)'\
                          % ((ch,) * 4),
            'dim': pb_def['dim'] - 1,
            'class': cb.CoefDimDim,
        },
    })

    requirements.update({
        'pis_' + ch: {
            'variables': ['p' + ch],
            'class': cb.ShapeDim,
        },
        'corrs_one' + ch: {
            'variable': ['pM'],
            'ebcs': ebcs_eta[ch],
            'epbcs': [],
            'save_name': 'corrs_%s_one%s' % (pb_def['name'], ch),
            'class': cb.CorrSetBCS,
        },
        'corrs_eta' + ch: {
            'ebcs': ebcs_eta[ch],
            'epbcs': all_periodicYM,
            'equations': {
                'eq_eta': 'dw_diffusion.iV.YM(mat2M.k, qM, pM) = 0',
                },
            'class': cb.CorrOne,
            'save_name': 'corrs_%s_eta%s' % (pb_def['name'], ch),
        },
        'corrs_pi' + ch: {
            'requires': ['pis_' + ch],
            'set_variables': [('Pi_' + ch, 'pis_' + ch, 'p' + ch)],
            'ebcs': [],
            'epbcs': all_periodicY[ch],
            'lcbcs': ['imv' + ch],
            'equations': {
                'eq_pi': """dw_diffusion.iV.Y%s(mat2%s.k, q%s, p%s)
                            + dw_volume_dot.iV.Y%s(q%s, ls%s)
                            = - dw_diffusion.iV.Y%s(mat2%s.k, q%s, Pi_%s)"""
                            % ((ch,) * 11),
                'eq_imv': 'dw_volume_dot.iV.Y%s(lv%s, p%s) = 0' % ((ch,) * 3),
            },
            'dim': pb_def['dim'] - 1,
            'class': cb.CorrDim,
            'save_name': 'corrs_%s_pi%s' % (pb_def['name'], ch),
        },
    })

    for ipm in ['p', 'm']:
        coefs.update({
            'E' + ipm + ch: {  # test+
                'requires': ['corrs_eta' + ch],
                'set_variables': [('corr_M', 'corrs_eta' + ch, 'pM')],
                'expression': 'ev_surface_integrate.iS.bYM%s(corr_M)' % ipm,
                'set_volume': 'bYp',
                'class': cb.CoefOne,
            },
            'F' + ipm + ch: {  # test+
                'requires': ['corrs_one' + ch, 'corrs_gamma_' + ipm],
                'set_variables': [('corr1_M', 'corrs_one' + ch, 'pM'),
                                  ('corr2_M', 'corrs_gamma_' + ipm, 'pM')],
                'expression': """dw_diffusion.iV.YM(mat2M.k, corr1_M, corr2_M)
                          - %e * ev_surface_integrate.iS.bYM%s(corr1_M)"""\
                                 % (1.0/param_h, ipm),
                'class': cb.CoefOne,
            },
        })

    for i_io in range(len(val['io_nd_grp'])):
        io = '%s_%d' % (ch, i_io + 1)

        coefs.update({
            'S' + io: {  # [Rohan1] (4.28), test+
                'requires': ['corrs_gamma_' + io, 'pis_' + ch],
                'set_variables': set_corr_S,
                'expression': 'dw_diffusion.iV.Y%s(mat2%s.k,corr1_%s,corr2_%s)'
                              % ((ch,) * 4),
                'dim': pb_def['dim'] - 1,
                'class': cb.CoefDim,
            },
            'P' + io: {  # test+
                'requires': ['pis_' + ch, 'corrs_pi' + ch],
                'set_variables': set_corr_cc,
                'expression': 'ev_surface_integrate.iS.bY%s(corr1_%s)'\
                              % (io, ch),
                'set_volume': 'bYp',
                'dim': pb_def['dim'] - 1,
                'class': cb.CoefDim,
            },
            'S_test' + io: {
                'requires': ['corrs_pi' + ch],
                'set_variables': [('corr1_' + ch, 'corrs_pi' + ch, 'p' + ch)],
                'expression': '%e * ev_surface_integrate.iS.bY%s(corr1_%s)'\
                              % (1.0 / param_h, io, ch),
                'dim': pb_def['dim'] - 1,
                'class': cb.CoefDim,
            },
        })

        requirements.update({
            'corrs_gamma_' + io: {
                'requires': [],
                'variables': ['p' + ch, 'q' + ch],
                'ebcs': [],
                'epbcs': all_periodicY[ch],
                'lcbcs': ['imv' + ch],
                'equations': {
                    'eq_gamma': """dw_diffusion.iV.Y%s(mat2%s.k, q%s, p%s)
                                   + dw_volume_dot.iV.Y%s(q%s, ls%s)
                                   = %e * dw_surface_integrate.iS.bY%s(q%s)"""
                                    % ((ch,) * 7 + (1.0/param_h, io, ch)),
                    'eq_imv': 'dw_volume_dot.iV.Y%s(lv%s, p%s) = 0'
                              % ((ch,) * 3),
                },
                'class': cb.CorrOne,
                'save_name': 'corrs_%s_gamma_%s' % (pb_def['name'], io),
            },
        })

        for i_io2 in range(len(val['io_nd_grp'])):
            io2 = '%s_%d' % (ch, i_io2 + 1)
            io12 = '%s_%d' % (io, i_io2 + 1)
            coefs.update({
                'R' + io12: {  # test+
                    'requires': ['corrs_gamma_' + io2],
                    'set_variables': [('corr1_' + ch, 'corrs_gamma_' + io2,
                                       'p' + ch)],
                    'expression': 'ev_surface_integrate.iS.bY%s(corr1_%s)'\
                                  % (io, ch),
                    'set_volume': 'bYp',
                    'class': cb.CoefOne,
                },
            })

solvers = {
    'ls': ('ls.scipy_direct', {}),
    'newton': ('nls.newton', {
        'i_max': 1,
    })
}



"""
Multiprocessing functions.
"""
import logging
import os
import six

try:
    from mpi4py import MPI
    mpi_comm = MPI.COMM_WORLD
    mpi_rank = mpi_comm.Get_rank()
    mpi_status = MPI.Status()
    use_multiprocessing = mpi_comm.Get_size() > 1
except:
    use_multiprocessing = False


global_multiproc_dict = {}
mpi_master = 0


class MPILogFile(MPI.File):
    def write(self, *args, **kwargs):
        self.Write_shared(*args, **kwargs)


class MPIFileHandler(logging.FileHandler):
    "MPI file class for logging process communication."
    def __init__(self, filename, mode=MPI.MODE_WRONLY,
                 encoding=None, delay=0, comm=MPI.COMM_WORLD):
        self.baseFilename = os.path.abspath(filename)
        self.mode = mode
        self.encoding = encoding
        self.comm = comm
        if delay:
            logging.Handler.__init__(self)
            self.stream = None
        else:
            logging.StreamHandler.__init__(self, self._open())

    def _open(self):
        stream = MPILogFile.Open(self.comm, self.baseFilename, self.mode)
        stream.Set_atomicity(True)
        return stream

    def close(self):
        if self.stream:
            self.stream.Sync()
            self.stream.Close()
            self.stream = None


def set_logging_level(log_level='info'):
    if log_level == 'debug':
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)


def get_logger(log_filename='multiproc_mpi.log'):
    """Get the MPI logger which log information into a shared file."""
    open(log_filename, 'w').close()  # empy log file

    log_id = 'master' if mpi_rank == 0 else 'slave%d' % mpi_comm.rank
    logger = logging.getLogger(log_id)
    logger.setLevel(logging.INFO)

    mh = MPIFileHandler(log_filename)
    formatter = logging.Formatter('%(asctime)s - %(levelname)s'
                                  + ' - %(name)s: %(message)s')
    mh.setFormatter(formatter)
    logger.addHandler(mh)

    return logger


logger = get_logger()


def enum(*sequential):
    enums = dict(zip(sequential, range(len(sequential))))
    reverse = dict((value, key) for key, value in six.iteritems(enums))
    enums['name'] = reverse
    return type('Enum', (), enums)


tags = enum('READY', 'DONE', 'START', 'CONTINUE',
            'LOCK', 'UNLOCK',
            'SET_DICT', 'SET_DICT_IMMUTABLE', 'GET_DICT', 'DICT_VAL',
            'SET_DICT_STATUS', 'GET_DICT_KEYS', 'DICT_KEYS',
            'GET_DICT_LEN', 'DICT_LEN', 'GET_DICT_IN', 'DICT_IN',
            'GET_QUEUE', 'PUT_QUEUE', 'QUEUE_VAL')


def cpu_count():
    """Get the number of MPI nodes."""
    return mpi_comm.Get_size()


def get_int_value(name, init_value=0):
    """Get the remote integer value."""
    key = '__int_%s' % name
    rdict = get_dict(key, mutable=True)
    if mpi_rank == mpi_master:
        val = RemoteInt(rdict, init_value)
    else:
        val = RemoteInt(rdict)

    return val


def get_queue(name):
    """Get the queue."""
    key = '__queue_%s' % name
    if mpi_rank == mpi_master:
        if key not in global_multiproc_dict:
            queue = RemoteQueueMaster(name)
            global_multiproc_dict[key] = queue
        else:
            queue = global_multiproc_dict[key]
            queue.clean()
    else:
        queue = RemoteQueue(name)

    return queue


def get_dict(name, mutable=False, clear=False, soft_set=False):
    """Get the remote dictionary."""
    if mpi_rank == mpi_master:
        if name in global_multiproc_dict:
            if clear:
                global_multiproc_dict[name].clear()
                logger.info("cleaning dict %s" % name)
            else:
                logger.info("using existing dict %s" % name)
            return global_multiproc_dict[name]
        else:
            rdict = RemoteDictMaster(name, mutable=mutable, soft_set=soft_set)
            global_multiproc_dict[name] = rdict
            logger.info("new dict %s" % name)
            return rdict
    else:
        return RemoteDict(name, mutable=mutable)


class RemoteInt(object):
    """Remote intiger class, data saved in RemoteDict."""
    class IntDesc(object):
        def __get__(self, instance, owner=None):
            return instance.remote_dict['value']

        def __set__(self, instance, val):
            instance.remote_dict['value'] = val

    value = IntDesc()

    def __init__(self, remote_dict, value=None):
        self.remote_dict = remote_dict
        if value is not None:
            self.value = value


class RemoteQueueMaster(list):
    """Remote queue class - master side."""
    def __init__(self, name, mode='fifo', *args):
        list.__init__(self, *args)
        self.name = name
        self.mode = mode

    @staticmethod
    def get_gdict_key(name):
        return '__queue_%s' % name

    def get(self):
        if len(self) == 0:
            return None
        else:
            if self.mode == 'fifo':
                out = self.pop(0)
            elif self.mode == 'lifo':
                out = self.pop(-1)

            return out

    def put(self, value):
        self.append(value)

    def clean(self):
        del self[:]

    def remote_put(self, value, slave):
        self.put(value)

    def remote_get(self, slave):
        mpi_comm.isend(self.get(), dest=slave, tag=tags.QUEUE_VAL)
        logger.debug('sent %s to %d (%s)'
                     % (tags.name[tags.QUEUE_VAL], slave, self.name))


class RemoteQueue(object):
    """Remote queue class - slave side."""
    def __init__(self, name):
        self.name = name

    def get(self):
        mpi_comm.isend(self.name, dest=mpi_master, tag=tags.GET_QUEUE)
        logger.debug('sent %s to %d (%s)'
                     % (tags.name[tags.GET_QUEUE], mpi_master, self.name))
        value = mpi_comm.recv(source=mpi_master, tag=tags.QUEUE_VAL)
        logger.debug('received %s from %d'
                     % (tags.name[tags.QUEUE_VAL], mpi_master))

        return value

    def put(self, value):
        mpi_comm.isend((self.name, value), dest=mpi_master, tag=tags.PUT_QUEUE)
        logger.debug('sent %s to %d (%s)'
                     % (tags.name[tags.PUT_QUEUE], mpi_master, self.name))


class RemoteDictMaster(dict):
    """Remote dictionary class - master side."""
    def __init__(self, name, mutable=False, soft_set=False, *args):
        dict.__init__(self, *args)
        self.name = name
        self.mutable = mutable
        self.immutable_soft_set = soft_set

    def remote_set(self, data, slave, mutable=False):
        key, value = data
        if not self.mutable and key in self:
            if self.immutable_soft_set:
                mpi_comm.isend(False, dest=slave, tag=tags.SET_DICT_STATUS)
            else:
                msg = "imutable dict '%s'! key '%s' already in global dict"\
                    % (self.name, key)
                logger.error(msg)
                mpi_comm.isend(False, dest=slave, tag=tags.SET_DICT_STATUS)
                raise(KeyError)
        else:
            self.__setitem__(key, value)
            logger.debug('set master dict (%s[%s])' % (self.name, key))
            mpi_comm.isend(True, dest=slave, tag=tags.SET_DICT_STATUS)

    def remote_get(self, key, slave):
        if key in self:
            mpi_comm.isend(self.__getitem__(key), dest=slave,
                           tag=tags.DICT_VAL)
            logger.debug('sent %s to %d (%s[%s])'
                         % (tags.name[tags.DICT_VAL], slave, self.name, key))
        else:
            mpi_comm.isend(None, dest=slave, tag=tags.DICT_VAL)
            logger.error('RemoteDict KeyError (%s[%s])' % (self.name, key))
            raise(KeyError)

    def remote_get_keys(self, slave):
        mpi_comm.isend(self.keys(), dest=slave, tag=tags.DICT_KEYS)
        logger.debug('sent %s to %d (%s)'
                     % (tags.name[tags.DICT_KEYS], slave, self.name))

    def remote_get_len(self, slave):
        mpi_comm.isend(self.__len__(), dest=slave, tag=tags.DICT_LEN)
        logger.debug('sent %s to %d (%s)'
                     % (tags.name[tags.DICT_LEN], slave, self.name))

    def remote_get_in(self, key, slave):
        mpi_comm.isend(self.__contains__(key), dest=slave, tag=tags.DICT_IN)
        logger.debug('sent %s to %d (%s)'
                     % (tags.name[tags.DICT_IN], slave, self.name))


class RemoteDict(object):
    """Remote dictionary class - slave side."""
    def __init__(self, name, mutable=False):
        self._dict = {}
        self.mutable = mutable
        self.name = name

    def _setitem(self, key, value, tag):
        mpi_comm.isend((self.name, key, value), dest=mpi_master, tag=tag)
        logger.debug('sent %s to %d (%s[%s])'
                     % (tags.name[tag], mpi_master, self.name, key))
        stat = mpi_comm.recv(source=mpi_master, tag=tags.SET_DICT_STATUS)
        logger.debug('recevied %s from %d'
                     % (tags.name[tags.SET_DICT_STATUS], mpi_master))

        return stat

    def __setitem__(self, key, value):
        if self.mutable:
            self._setitem(key, value, tags.SET_DICT)
            self._dict[key] = value
        else:
            if key in self._dict:
                msg = "imutable dict '%s'! key '%s' already in local dict"\
                    % (self.name, key)
                logger.error(msg)
                raise(KeyError)
            else:
                stat = self._setitem(key, value, tags.SET_DICT_IMMUTABLE)
                if stat:
                    self._dict[key] = value
                else:
                    msg = \
                        "imutable dict '%s'! key '%s' already in global dict"\
                        % (self.name, key)
                    logger.error(msg)
                    raise(KeyError)

    def __getitem__(self, key):
        if key in self._dict and not self.mutable:
            logger.debug('get value from local dict! (%s[%s])'
                         % (self.name, key))
        else:
            mpi_comm.isend((self.name, key), dest=mpi_master,
                           tag=tags.GET_DICT)
            logger.debug('sent %s to %d (%s)'
                         % (tags.name[tags.GET_DICT], mpi_master, key))
            data = mpi_comm.recv(source=mpi_master, tag=tags.DICT_VAL)
            logger.debug('received %s from %d' % (tags.name[tags.DICT_VAL],
                                                  mpi_master))
            if data is not None:
                self._dict[key] = data
            else:
                logger.error('RemoteDict KeyError (%s[%s])' % (self.name, key))
                raise(KeyError)

        return self._dict[key]

    def __len__(self):
        mpi_comm.isend((self.name,), dest=mpi_master, tag=tags.GET_DICT_LEN)
        length = mpi_comm.recv(source=mpi_master, tag=tags.DICT_LEN)
        return length

    def __contains__(self, key):
        if key in self._dict and not self.mutable:
            is_in = True
        else:
            mpi_comm.isend((self.name, key), dest=mpi_master,
                           tag=tags.GET_DICT_IN)
            is_in = mpi_comm.recv(source=mpi_master, tag=tags.DICT_IN)

        return is_in

    def keys(self):
        mpi_comm.isend((self.name,), dest=mpi_master, tag=tags.GET_DICT_KEYS)
        keys = mpi_comm.recv(source=mpi_master, tag=tags.DICT_KEYS)
        return keys

    def get(self, key, default=None):
        if key in self.keys():
            return self.__getitem__(key)
        else:
            return default

    def update(self, other):
        for k in other.keys():
            self.__setitem__(k, other[k])


def is_remote_dict(d):
    """Return True if 'd' is RemoteDict or RemoteDictMaster instance."""
    return isinstance(d, RemoteDict) or isinstance(d, RemoteDictMaster)


class RemoteLock(object):
    """Remote lock class - lock and unlock restricted access to the master."""
    def __init__(self):
        self.locked = False

    def acquire(self):
        mpi_comm.isend(None, dest=mpi_master, tag=tags.LOCK)
        self.locked = True

    def release(self):
        mpi_comm.isend(None, dest=mpi_master, tag=tags.UNLOCK)
        self.locked = False


def wait_for_tag(wtag, num=1):
    ndone = num
    start = MPI.Wtime()
    while ndone > 0:
        mpi_comm.recv(source=MPI.ANY_SOURCE, tag=wtag, status=mpi_status)
        tag = mpi_status.Get_tag()
        source = mpi_status.Get_source()
        logger.debug('received %s from %d (%.03fs)' % (tags.name[tag],
                                                       source,
                                                       MPI.Wtime() - start))
        if tag == wtag:
            ndone -= 1


def slave_get_task(name=''):
    """Start the slave nodes."""
    mpi_comm.isend(mpi_rank, dest=mpi_master, tag=tags.READY)
    logger.debug('%s ready' % name)
    task, data = mpi_comm.bcast(None, root=mpi_master)
    logger.info('%s received task %s' % (name, task))

    return task, data


def slave_task_done(task=''):
    """Stop the slave nodes."""
    mpi_comm.isend(mpi_rank, dest=mpi_master, tag=tags.DONE)
    logger.info('%s stopped' % task)


def get_slaves():
    """Get the list of slave nodes"""
    slaves = range(mpi_comm.Get_size())
    slaves.remove(mpi_master)
    return slaves


def master_send_task(task, data):
    """Send task to all slaves."""
    slaves = get_slaves()
    wait_for_tag(tags.READY, len(slaves))
    logger.info('all nodes are ready for task %s' % task)
    mpi_comm.bcast((task, data), root=mpi_master)


def master_send_continue():
    """Send 'continue' to all slaves."""
    for ii in get_slaves():
        mpi_comm.send(None, dest=ii, tag=tags.CONTINUE)
    logger.info('slave nodes - continue')


def master_loop():
    """Run the master loop - wait for requests from slaves."""
    logger.info('main loop started')
    master_send_task('calculate', None)

    ndone = len(get_slaves())
    source = MPI.ANY_SOURCE
    while ndone > 0:
        data = mpi_comm.recv(source=source, tag=MPI.ANY_TAG, status=mpi_status)
        tag = mpi_status.Get_tag()
        slave = mpi_status.Get_source()
        logger.debug('received %s from %d' % (tags.name[tag], slave))
        if tag == tags.DONE:
            ndone -= 1
        elif tag == tags.LOCK:
            source = slave
        elif tag == tags.UNLOCK:
            source = MPI.ANY_SOURCE
        elif tag == tags.SET_DICT:
            global_multiproc_dict[data[0]].remote_set(data[1:], slave,
                                                      mutable=True)
        elif tag == tags.SET_DICT_IMMUTABLE:
            global_multiproc_dict[data[0]].remote_set(data[1:], slave)
        elif tag == tags.GET_DICT:
            global_multiproc_dict[data[0]].remote_get(data[1], slave)
        elif tag == tags.GET_DICT_KEYS:
            global_multiproc_dict[data[0]].remote_get_keys(slave)
        elif tag == tags.GET_DICT_LEN:
            global_multiproc_dict[data[0]].remote_get_len(slave)
        elif tag == tags.GET_DICT_IN:
            global_multiproc_dict[data[0]].remote_get_in(data[1], slave)
        elif tag == tags.GET_QUEUE:
            qkey = RemoteQueueMaster.get_gdict_key(data)
            global_multiproc_dict[qkey].remote_get(slave)
        elif tag == tags.PUT_QUEUE:
            qkey = RemoteQueueMaster.get_gdict_key(data[0])
            global_multiproc_dict[qkey].remote_put(data[1], slave)

    logger.info('main loop finished')
